@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -145,6 +145,9 @@ EXTERN char *dosemu_midi_path INIT("~/" LOCALDIR_BASE_NAME "/run/" DOSEMU_MIDI);
 #undef X2_SUPPORT
 #undef X_GRAPHICS
 
+/* Define this to use the old Slang keyboard code */
+#undef USE_OLD_SLANG_KBD
+
 /*
  *  DEFINE this, if you have joystick support in the kernel
  *  or have the joystick module loaded (needing <linux/joystick.h>)
@@ -181,9 +184,16 @@ EXTERN char *dosemu_midi_path INIT("~/" LOCALDIR_BASE_NAME "/run/" DOSEMU_MIDI);
 /* Define this to use the new interrupt handling code */
 #undef USE_NEW_INT
 
+/* Define this for ASPI (generic SCSI) support */
+#undef ASPI_SUPPORT
+
 /* Define this to use the X86 CPU emulator */
 #undef X86_EMULATOR
 
+/* Define this on glibc >= 2.1,
+ * because though it has llseek in the library, it doesn't prototype it */
+#undef NEED_LLSEEK_PROTOTYPE
+
 @BOTTOM@
 
 #endif /* CONFIG_H */
