@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -45,8 +45,8 @@ extern int errno;
 #endif
 
 #ifdef X86_EMULATOR
-extern int e_sig_pending;
-#define E_SIGNAL	if (config.cpuemu>1) e_sig_pending=1
+#include "cpu-emu.h"
+#define E_SIGNAL	if (config.cpuemu>1) CEmuStat|=CeS_SIGPEND
 #else
 #define E_SIGNAL
 #endif
@@ -65,7 +65,7 @@ static struct SIGNAL_queue signal_queue[MAX_SIG_QUEUE_SIZE];
 static sigset_t oldset;
 
 #ifdef __linux__
-#if __GLIBC__ > 1		/* glibc-2 version */
+#if GLIBC_VERSION_CODE >= 2000
 /*
  * Thomas Winder <thomas.winder@sea.ericsson.se> wrote:
  * glibc-2 uses a different struct sigaction type than the one used in
@@ -276,6 +276,10 @@ sti(void)
  */
 void handle_signals(void) {
   if ( SIGNAL_head != SIGNAL_tail ) {
+#ifdef X86_EMULATOR
+    if ((config.cpuemu>1) && (d.emu>3))
+      {e_printf("EMU86: SIGNAL at %d\n",SIGNAL_head);}
+#endif
     signal_queue[SIGNAL_head].signal_handler();
     SIGNAL_head = (SIGNAL_head + 1) % MAX_SIG_QUEUE_SIZE;
 /* 
@@ -472,7 +476,7 @@ void SIGALRM_call(void)
   /* this should be for per-second activities, it is actually at
    * 200ms more or less (PARTIALS=5) */
   if ((pic_sys_time-cnt200) >= (PIT_TICK_RATE/PARTIALS)) {
-    cnt200 += (PIT_TICK_RATE/PARTIALS);
+    cnt200 = pic_sys_time;
 /*    g_printf("**** ALRM: %dms\n",(1000/PARTIALS)); */
 
 #ifdef IPX
@@ -577,17 +581,18 @@ sigalrm(int sig, struct sigcontext_struct context)
     dpmi_sigio(&context);
   SIGNAL_save(SIGALRM_call);
 }
-#endif
 
 #ifdef X86_EMULATOR
+/* this is the same thing, but with a pointer parameter */
 void
-e_sigalrm(void)
+e_sigalrm(struct sigcontext_struct *context)
 {
-/*  if (in_dpmi && !in_vm86)
-    dpmi_sigio(&context); */
+  if (in_dpmi && !in_vm86)
+    dpmi_sigio(context);
   SIGNAL_save(SIGALRM_call);
 }
 #endif
+#endif
 
 #ifdef __NetBSD__
 void
