@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -45,6 +45,7 @@
 #include "keyb_server.h"
 
 #define DJGPP_HACK	/* AV Feb 97 */
+#undef  DEBUG_INT1A
 
 #if X_GRAPHICS
 /* prototype is in X.h -- 1998/03/08 sw */
@@ -63,6 +64,9 @@ static unsigned long  precard_eip, precard_cs;
 
 static struct timeval scr_tv;        /* For translating UNIX <-> DOS times */
 
+/* set if some directories are mounted during startup */
+int redir_state = 0;
+
 #ifdef USE_MRP_JOYSTICK
 #include <linux/joystick.h>
 #include <fcntl.h>
@@ -216,7 +220,7 @@ static int dos_helper(void)
     show_ints(HI(bx), LO(bx));
     break;
 
-  case DOS_HELPER_PRINT_STRING:	/* SHOW INTS, BH-BL */
+  case DOS_HELPER_PRINT_STRING:	/* PRINT STRING ES:DX */
     g_printf("DOS likes us to print a string\n");
     ds_printf("DOS to EMU: \"%s\"\n",SEG_ADR((char *), es, dx));
     break;
@@ -345,6 +349,8 @@ static int dos_helper(void)
     }
     v_printf("Restore hi vector area\n");
     MEMCPY_2DOS(0x380,save_hi_ints,128);
+    config.emuretrace <<= 1;
+    emu_video_retrace_on();
     break;
 
   case DOS_HELPER_GET_DEBUG_STRING:
@@ -423,6 +429,18 @@ static int dos_helper(void)
       break;
     }
 
+  case DOS_HELPER_ASPI_HELPER: {
+#ifdef ASPI_SUPPORT
+      extern void aspi_helper(int);
+      A_printf("ASPI: in 0x41 handler! ax=0x%04x, bx=0x%04x, dx=0x%04x, "
+           "cx=0x%04x\n", LWORD(eax), LWORD(ebx), LWORD(edx), LWORD(ecx));
+      aspi_helper(HI(ax));
+#else
+      LWORD(eax) = 0;
+#endif
+      break;
+  }
+
   case DOS_HELPER_RUN_UNIX:
     g_printf("Running Unix Command\n");
     run_unix_command(SEG_ADR((char *), es, dx));
@@ -441,16 +459,33 @@ static int dos_helper(void)
     break;   
 
   case DOS_HELPER_0x53:
-    if (config.secure) {
-      CARRY;
-      LWORD(eax) = 0;
-      break;
-    }
     {
-	PRIV_SAVE_AREA
-	enter_priv_off();
-	LWORD(eax) = system(SEG_ADR((char *), es, dx));
-	leave_priv_setting();
+        extern int run_system_command(char *);
+        LWORD(eax) = run_system_command(SEG_ADR((char *), es, dx));
+	break;
+    }
+
+  case DOS_HELPER_GET_CPU_SPEED:
+    {
+	if (config.rdtsc)
+		REG(eax) = (LLF_US << 16) / config.cpu_spd;
+	else	REG(eax) = 0;
+	break;
+    }
+
+  case DOS_HELPER_GET_TERM_TYPE:
+    {
+	int i;
+
+	/* NOTE: we assume terminal/video init has completed before coming here */
+	if (config.X) i = 2;			/* X keyboard */
+	else if (config.console_keyb) i = 0;	/* raw keyboard */
+	else i = 1;				/* Slang keyboard */
+
+	if (config.console_video) i |= 0x10;
+	if (config.graphics)      i |= 0x20;
+	if (config.dualmon)       i |= 0x40;
+	LWORD(eax) = i;
 	break;
     }
 
@@ -476,10 +511,15 @@ static int dos_helper(void)
         break;
 #ifdef X86_EMULATOR
   case DOS_HELPER_CPUEMUON:
-  	if (config.cpuemu && !in_dpmi) enter_cpu_emu();
+#ifdef DONT_DEBUG_BOOT
+	memcpy(&d,&d_save,sizeof(struct debug_flags));
+#endif
+	/* we could also enter from inside dpmi, provided we already
+	 * mirrored the LDT into the emu's own one */
+  	if ((config.cpuemu==1) && !in_dpmi) enter_cpu_emu();
         break;
   case DOS_HELPER_CPUEMUOFF:
-  	if (config.cpuemu && !in_dpmi) leave_cpu_emu();
+  	if ((config.cpuemu>1) && !in_dpmi) leave_cpu_emu();
         break;
 #endif
     case DOS_HELPER_XCONFIG:
@@ -508,7 +548,8 @@ static int dos_helper(void)
     }
     break;
 
-
+  /* here we try to hook a possible plugin */
+  #include "plugin_inte6.h"
 
   default:
     error("bad dos helper function: AX=0x%04x\n", LWORD(eax));
@@ -645,7 +686,7 @@ static void int15(u_char i)
   case 0xc0:
     LWORD(es) = ROM_CONFIG_SEG;
     LWORD(ebx) = ROM_CONFIG_OFF;
-    LO(ax) = 0;
+    HI(ax) = 0;
     return;
   case 0xc1:
     CARRY;
@@ -779,7 +820,6 @@ void set_ticks(unsigned long new)
 
 static void int1a(u_char i)
 {
-  long t;
   time_t time_val;
   struct timeval;
   struct timezone;
@@ -803,21 +843,74 @@ Notes:	there are approximately 18.2 clock ticks per second, 1800B0h per 24 hrs
 ->	  midnight flag and will fail to advance the date
 */
   case 0:			/* read time counter */
-    /* it works because pic_sys_time has a zero reference. It doesn't
-     * account for timer speedups, but any decent DOS program should
-     * call back the original int8 at the 18.2 Hz rate... */
-    last_ticks = sys_base_ticks + usr_delta_ticks + (pic_sys_time >> 16);
 
-    /* has the midnight passed? */
-    if (last_ticks > 1573040) {
-      *(u_char *) (TICK_OVERFLOW_ADDR) += 0x1;
-      last_ticks -= 1573040;
+#define BIOSTIMER_ONLY_VIEW
+#ifdef BIOSTIMER_ONLY_VIEW
+
+    /* We rely on the INT8 routine doing the right thing,
+     * DOS apps too rely on the relationship between INT1A and 0x46c timer.
+     * We already do all appropriate things to trigger the simulated INT8
+     * correctly (well, sometimes faking it), so the 0x46c timer incremented
+     * by the (realmode) INT8 handler should be always in sync.
+     * Therefore, we keep INT1A,AH0 simple instead of trying to be too clever;-)
+     */
+    {
+      static int first = 1;
+      if (first) {
+        /* take over the correct value _once_ only */
+        *((unsigned long *)(BIOS_TICK_ADDR)) =
+               (unsigned long)(pic_sys_time >> 16)
+             + (sys_base_ticks + usr_delta_ticks);
+        first = 0;
+      }
     }
-    LO(ax) = *(u_char *) (TICK_OVERFLOW_ADDR);
+    last_ticks = *((unsigned long *)(BIOS_TICK_ADDR));
+    LO(ax) = *((u_char *)(TICK_OVERFLOW_ADDR));
     LWORD(ecx) = (last_ticks >> 16) & 0xffff;
     LWORD(edx) = last_ticks & 0xffff;
-    g_printf("TIMER: read timer = %lu\n", last_ticks);
+    *((u_char *)(TICK_OVERFLOW_ADDR)) = 0; /* clear the midnight flag */
+
+#else /* not BIOSTIMER_ONLY_VIEW */
+
+    /* pic_sys_time is a zero-based tick(1.19MHz) counter. As such, if we
+     * shift it right by 16 we get the number of PIT0 overflows, that is,
+     * the number of 18.2ms timer ticks elapsed since starting dosemu. This
+     * is independent of any int8 speedup a program can set, since the PIT0
+     * counting frequency is fixed. The count overflows after 7 1/2 years.
+     * usr_delta_ticks is 0 as long as nobody sets a new time (B-1A01)
+     */
+#ifdef DEBUG_INT1A
+    g_printf("TIMER: sys_base_ticks=%ld usr_delta_ticks=%ld pic_sys_time=%#Lx\n",
+    	sys_base_ticks, usr_delta_ticks, pic_sys_time);
+#endif
+    last_ticks = (unsigned long)(pic_sys_time >> 16);
+    last_ticks += (sys_base_ticks + usr_delta_ticks);
+
+    /* has the midnight passed? */
+    if (last_ticks > TICKS_IN_A_DAY) {
+      *((u_char *)(TICK_OVERFLOW_ADDR)) |= 0x1;
+      last_ticks -= TICKS_IN_A_DAY;
+      /* since pic_sys_time continues to increase, avoid further midnight
+       * overflows */
+      sys_base_ticks -= TICKS_IN_A_DAY;
+    }
+    LO(ax) = *((u_char *)(TICK_OVERFLOW_ADDR));
+    LWORD(ecx) = (last_ticks >> 16) & 0xffff;
+    LWORD(edx) = last_ticks & 0xffff;
+
+#ifdef DEBUG_INT1A
+    if (d.general) {
+      long k = last_ticks/18.2065;	/* sorry */
+      time(&time_val);
+      g_printf("INT1A: read timer = %ld (%ld:%ld:%ld) %s\n", last_ticks,
+      	k/3600, (k%3600)/60, (k%60), ctime(&time_val));
+    }
+#else
+    g_printf("INT1A: read timer=%ld, midnight=%d\n", last_ticks, LO(ax));
+#endif
+
     set_ticks(last_ticks);	/* set_ticks is in rtc.c */
+#endif /* not BIOSTIMER_ONLY_VIEW */
     break;
 
 /*
@@ -832,13 +925,28 @@ Notes:	there are approximately 18.2 clock ticks per second, 1800B0h per 24 hrs
 SeeAlso: AH=00h,AH=03h,INT 21/AH=2Dh
 */
   case 1:			/* write time counter */
-    t = sys_base_ticks + (pic_sys_time >> 16);
-    last_ticks = (LWORD(ecx) << 16) | (LWORD(edx) & 0xffff);
-    usr_delta_ticks = last_ticks - t;
-    set_ticks(last_ticks);
-    g_printf("INT1A: set timer to %#lx\n", last_ticks);
-    break;
+    {
+      /* get current system time and check it (previous usr_delta could
+       * be != 0) */
+      long t;
+      do {
+        t = (pic_sys_time >> 16) + sys_base_ticks;
+        if (t < 0) sys_base_ticks += TICKS_IN_A_DAY;
+      } while (t < 0);
 
+      /* get user-requested time */
+      last_ticks = (LWORD(ecx) << 16) | (LWORD(edx) & 0xffff);
+
+      usr_delta_ticks = last_ticks - t;
+      *(u_char *) (TICK_OVERFLOW_ADDR) = 0;
+#ifdef DEBUG_INT1A
+      g_printf("TIMER: sys_base_ticks=%ld usr_delta_ticks=%ld pic_sys_time=%#Lx\n",
+    	sys_base_ticks, usr_delta_ticks, pic_sys_time);
+#endif
+      set_ticks(last_ticks);
+      g_printf("INT1A: set timer to %ld\n", last_ticks);
+      break;
+    }
 /*
 --------B-1A02-------------------------------
 INT 1A - TIME - GET REAL-TIME CLOCK TIME (AT,XT286,PS)
@@ -1033,7 +1141,7 @@ static int ms_dos(int nr)
     char *ptr = SEG_ADR((char *), ds, dx);
 
     /* ignore explicitly selected drive by incrementing ptr by 1 */
-    if (config.emubat && !strncmp(ptr + 1, ":\\AUTOEXEC.BAT", 14)) {
+    if (config.emubat && !strncasecmp(ptr + 1, ":\\AUTOEXEC.BAT", 14)) {
       ext_fix(config.emubat);
       sprintf(ptr + 1, ":\\AUTOEXEC.%-3s", config.emubat);
       d_printf("DISK: Substituted %s for AUTOEXEC.BAT\n", ptr + 1);
@@ -1100,9 +1208,11 @@ static int ms_dos(int nr)
     return 0;
 
   default:
+#ifndef USE_NEW_INT
     if (!in_dpmi)
       g_printf("INT21 (0x%02x):  we shouldn't be here! ax=0x%04x, bx=0x%04x\n",
 	     nr, LWORD(eax), LWORD(ebx));
+#endif
     return 0;
   }
   return 1;
@@ -1388,11 +1498,153 @@ static void int19(u_char i) {
   boot();
 }
 
+
+/*
+ * Turn all simulated FAT devices into network drives.
+ */
+static void redirect_devices()
+{
+  extern int RedirectDisk(int, char *, int);
+
+  static char s[256] = "\\\\LINUX\\FS", *t = s + 10;
+  int i, j;
+
+  for (i = 0; i < MAX_HDISKS; i++) {
+    if(hdisktab[i].type == DIR_TYPE && hdisktab[i].fatfs) {
+      strncpy(t, hdisktab[i].dev_name, 245);
+      s[255] = 0;
+      j = RedirectDisk(i + 2, s, hdisktab[i].rdonly);
+
+      ds_printf("INT21: redirecting %c: %s (err = %d)\n", i + 'C', j ? "failed" : "ok", j);
+    }
+  }
+}
+
+/*
+ * Activate the redirector just before the first int 21h file open call.
+ *
+ * To use this feature, set redir_state = 1 and make sure int 21h is
+ * revectored.
+ */
+static int redir_it()
+{
+  static struct vm86_regs save_regs;
+  static unsigned x0, x1, x2, x3, x4;
+  unsigned u;
+
+  /*
+   * To start up the redirector we need (1) the list of list, (2) the DOS version and
+   * (3) the swappable data area. To get these, we reuse the original file open call.
+   */
+  switch(redir_state) {
+    case 1:
+      if(HI(ax) == 0x3d) {
+        save_regs = REGS;
+        redir_state = 2;
+        LWORD(eip) -= 2;
+        LWORD(eax) = 0x5200;
+        default_interrupt(0x21);
+        ds_printf("INT21 +1 (%d) at %04x:%04x: AX=%04x, BX=%04x, CX=%04x, DX=%04x, DS=%04x, ES=%04x\n",
+          redir_state, LWORD(cs), LWORD(eip), LWORD(eax), LWORD(ebx), LWORD(ecx), LWORD(edx), LWORD(ds), LWORD(es));
+        return 1;
+      }
+      break;
+
+    case 2:
+      x0 = LWORD(ebx); x1 = REG(es);
+      redir_state = 3;
+      LWORD(eip) -= 2;
+      LWORD(eax) = 0x3000;
+      default_interrupt(0x21);
+      ds_printf("INT21 +2 (%d) at %04x:%04x: AX=%04x, BX=%04x, CX=%04x, DX=%04x, DS=%04x, ES=%04x\n",
+        redir_state, LWORD(cs), LWORD(eip), LWORD(eax), LWORD(ebx), LWORD(ecx), LWORD(edx), LWORD(ds), LWORD(es));
+      return 2;
+      break;
+
+    case 3:
+      x4 = LWORD(eax);
+      redir_state = 4;
+      LWORD(eip) -= 2;
+      LWORD(eax) = 0x5d06;
+      default_interrupt(0x21);
+      ds_printf("INT21 +3 (%d) at %04x:%04x: AX=%04x, BX=%04x, CX=%04x, DX=%04x, DS=%04x, ES=%04x\n",
+        redir_state, LWORD(cs), LWORD(eip), LWORD(eax), LWORD(ebx), LWORD(ecx), LWORD(edx), LWORD(ds), LWORD(es));
+      return 3;
+      break;
+
+    case 4:
+      x2 = LWORD(esi); x3 = REG(ds);
+      redir_state = 0;
+      u = x0 + (x1 << 4);
+      ds_printf("INT21: lol = 0x%x\n", u);
+      ds_printf("INT21: sda = 0x%x\n", x2 + (x3 << 4));
+      ds_printf("INT21: ver = 0x%02x\n", x4);
+
+      if(*(unsigned *) (u + 0x16)) {		/* Do we have a CDS entry? */
+        /* Init the redirector. */
+        LWORD(ecx) = x4;
+        LWORD(edx) = x0; REG(es) = x1;
+        LWORD(esi) = x2; REG(ds) = x3;
+        LWORD(ebx) = 0x500;
+        LWORD(eax) = 0x20;
+        mfs_inte6();
+
+        redirect_devices();
+      }
+      else {
+        ds_printf("INT21: this DOS has no CDS entry - redirector not used\n");
+      }
+
+      REGS = save_regs;
+      set_int21_revectored(-1);
+      break;
+  }
+
+  return 0;
+}
+
+
 /* MS-DOS */
 static void int21(u_char i) {
-  ds_printf("INT21 at %04x:%04x: AX=%04x, BX=%04x, CX=%04x, DX=%04x, DS=%04x, ES=%04x\n",
-       LWORD(cs), LWORD(eip),
+#ifdef X86_EMULATOR
+  static char buf[80];
+#endif
+  ds_printf("INT21 (%d) at %04x:%04x: AX=%04x, BX=%04x, CX=%04x, DX=%04x, DS=%04x, ES=%04x\n",
+       redir_state, LWORD(cs), LWORD(eip),
        LWORD(eax), LWORD(ebx), LWORD(ecx), LWORD(edx), LWORD(ds), LWORD(es));
+
+  if(redir_state && redir_it()) return;
+
+#if 1
+  if(HI(ax) == 0x3d) {
+    char *p = (char *) (((REG(ds)) << 4) + LWORD(edx));
+    int i;
+
+    ds_printf("INT21: open file \"");
+    for(i = 0; i < 64 && p[i]; i++) ds_printf("%c", p[i]);
+    ds_printf("\"\n");
+  }
+#endif
+
+#ifdef X86_EMULATOR
+  if ((HI(ax)==0x40) && LWORD(ecx)) {
+	char *dp = (char *)((LWORD(ds)<<4)+LWORD(edx));
+	unsigned int nb = LWORD(ecx);
+	if (nb>78) nb=78; memcpy(buf,dp,nb); buf[nb]=0;
+	ds_printf("WRITE: [%s]\n",buf);
+  }
+  else if (HI(ax)==9) {
+	char *dp = (char *)((LWORD(ds)<<4)+LWORD(edx));
+	char *q = buf;
+	int nb;
+	for (nb=0; (nb<78)&&(*dp!='$'); nb++) *q++ = *dp++;
+	buf[nb]=0;
+	ds_printf("WRITE: [%s]\n",buf);
+  }
+  else if ((HI(ax)==6) && (LO(ax)!=0xff)) {
+	ds_printf("WRITE: [%c]\n",isprint(LO(ax)? LO(ax):'.'));
+  }
+#endif
   if (!ms_dos(HI(ax)))
     default_interrupt(i);
 }
@@ -1415,7 +1667,7 @@ static void int23(u_char i)
 
 static void mouse_post_boot(void)
 {
-	void bios_f000(), bios_f000_int10_old();
+	extern void bios_f000_int10_old();
 	us *ptr;
 	/* This is needed here to revectoring the interrupt, after dos
 	 * has revectored it. --EB 1 Nov 1997 */
@@ -1485,6 +1737,11 @@ static void int29(u_char i) {
 
 static void int2f(u_char i)
 {
+#if 1
+  ds_printf("INT2F at %04x:%04x: AX=%04x, BX=%04x, CX=%04x, DX=%04x, DS=%04x, ES=%04x\n",
+       LWORD(cs), LWORD(eip),
+       LWORD(eax), LWORD(ebx), LWORD(ecx), LWORD(edx), LWORD(ds), LWORD(es));
+#endif
   switch (LWORD(eax)) {
   case INT2F_IDLE_MAGIC:   /* magic "give up time slice" value */
     if (config.hogthreshold)
@@ -1498,63 +1755,39 @@ static void int2f(u_char i)
       return;
     break;
 #endif
-      
-  case 0x1687:            /* Call for getting DPMI entry point */
-    dpmi_get_entry_point();
-    return;
-
-  case 0x1683:
-    LWORD (ebx) = 0;		/* W95: number of virtual machine */
-  case 0x1681:		/* W95: enter critical section */
-    if (in_dpmi && in_win31) {
-	D_printf ("WIN: enter critical section\n");
-	/* LWORD(eax) = 0;	W95 DDK says no return value */
-	return;
     }
+
+  switch (HI(ax)) {
+  case 0x11:              /* redirector call? */
+    if (mfs_redirector())
+    return;
     break;
-  case 0x1682:		/* W95: exit critical section */
-    if (in_dpmi && in_win31) {
-	D_printf ("WIN: exit critical section\n");
-	/* LWORD(eax) = 0;	W95 DDK says no return value */
-	return;
+
+  case 0x16:		/* misc PM/Win functions */
+    if (!config.dpmi) {
+/*  d.emu=4; */
+      break;		/* fall into default_interrupt() */
     }
-    break;
-    return;
-
-  case 0x1686:            /* Are we in protected mode? */
-    D_printf("DPMI CPU mode check in real mode.\n");
-    if (in_dpmi) /* set AX to zero only if program executes in protected mode */
-    LWORD(eax) = 0;	/* say ok */
-		 /* else let AX untouched (non-zero) */
-    return;
-
-  case 0x1600:		/* WINDOWS ENHANCED MODE INSTALLATION CHECK */
+    switch (LO(ax)) {
+      case 0x00:		/* WINDOWS ENHANCED MODE INSTALLATION CHECK */
 #if 1			/* it seens this confuse winos2 */
     if (in_dpmi && in_win31) {
       D_printf("WIN: WINDOWS ENHANCED MODE INSTALLATION CHECK\n");
       LWORD(eax) = 0x0a03;	/* let's try enhaced mode 3.1 :-))))))) */
       return;
-    } else
-      break;
+      }
 #endif    
     break;
-  case 0x1605:		/* Win95 Initialization Notification */
+
+      case 0x05:		/* Win95 Initialization Notification */
     LWORD(ecx) = 0xffff;	/* say it`s NOT ok to run under Win */
-  case 0x1606:		/* Win95 Termination Notification */
-  case 0x1607:		/* Win95 Device CallOut */
-  case 0x1608:		/* Win95 Init Complete Notification */
-  case 0x1609:		/* Win95 Begin Exit Notification */
-    return;
-  case 0x1684:		/* Win95 Get Device Entry Point */
-    LWORD(edi) = 0;
-    WRITE_SEG_REG(es, 0);	/* say NO to Win95 ;-) */
-    return;
-  case 0x1685:		/* Win95 Switch VM + Call Back */
-    CARRY;
-    LWORD(eax) = 1;
+      case 0x06:		/* Win95 Termination Notification */
+      case 0x07:		/* Win95 Device CallOut */
+      case 0x08:		/* Win95 Init Complete Notification */
+      case 0x09:		/* Win95 Begin Exit Notification */
     return;
 
-  case 0x160a:			/* IDENTIFY WINDOWS VERSION AND TYPE */
+      case 0x0a:			/* IDENTIFY WINDOWS VERSION AND TYPE */
     if(in_dpmi && in_win31) {
       D_printf ("WIN: WINDOWS VERSION AND TYPE\n");
       LWORD(eax) =0;
@@ -1565,14 +1798,47 @@ static void int2f(u_char i)
       LWORD(ecx) = 0x0002;	/* standard mode */
 #endif      
       return;
-    } else
+        }
       break;
-  }
 
-  switch (HI(ax)) {
-  case 0x11:              /* redirector call? */
-    if (mfs_redirector())
+      case 0x83:
+        if (in_dpmi && in_win31)
+            LWORD (ebx) = 0;	/* W95: number of virtual machine */
+      case 0x81:		/* W95: enter critical section */
+        if (in_dpmi && in_win31) {
+	    D_printf ("WIN: enter critical section\n");
+	    /* LWORD(eax) = 0;	W95 DDK says no return value */
+	    return;
+  }
+      break;
+      case 0x82:		/* W95: exit critical section */
+        if (in_dpmi && in_win31) {
+	    D_printf ("WIN: exit critical section\n");
+	    /* LWORD(eax) = 0;	W95 DDK says no return value */
+	    return;
+  }
+        break;
+
+      case 0x84:		/* Win95 Get Device Entry Point */
+        LWORD(edi) = 0;
+        WRITE_SEG_REG(es, 0);	/* say NO to Win95 ;-) */
+        return;
+      case 0x85:		/* Win95 Switch VM + Call Back */
+        CARRY;
+        LWORD(eax) = 1;
+        return;
+
+      case 0x86:            /* Are we in protected mode? */
+        D_printf("DPMI CPU mode check in real mode.\n");
+        if (in_dpmi) /* set AX to zero only if program executes in protected mode */
+	    LWORD(eax) = 0;	/* say ok */
+		 /* else let AX untouched (non-zero) */
       return;
+
+      case 0x87:            /* Call for getting DPMI entry point */
+	dpmi_get_entry_point();
+	return;
+    }
     break;
 
   case INT2F_XMS_MAGIC:
@@ -2071,18 +2337,26 @@ void setup_interrupts(void) {
 #endif /* not USE_NEW_INT */
 }
 
-static struct revectored_struct saved_int21_revectored = {{0,0,0,0,0,0,0,0}};
 
-void set_int21_revectored(int all)
+void set_int21_revectored(int a)
 {
-  static int last_all = 0;
-  if (all < 0)
-    all = last_all;
-  if (all > 0)
+  static int rv_all = 0;
+  int i;
+
+  ds_printf("INT21: rv_all: %d + %d = ", rv_all, a);
+
+  rv_all += a;
+
+  if(rv_all > 0) {
     memset(&vm86s.int21_revectored, 0xff, sizeof(vm86s.int21_revectored));
-  else
-    vm86s.int21_revectored = saved_int21_revectored;
-  last_all = all;
+  }
+  else {
+    memset(&vm86s.int21_revectored, 0x00, sizeof(vm86s.int21_revectored));
+    for(i = 0; i < 0x100; i++)
+      if(can_revector_int21(i) == REVECT) set_revectored(i, &vm86s.int21_revectored);
+  }
+
+  ds_printf("%d\n", rv_all);
 }
 
 
@@ -2109,11 +2383,7 @@ void int_vector_setup(void)
     if (can_revector(i)==REVECT && i!=0x21)
       set_revectored(i, &vm86s.int_revectored);
 
-  for (i=0; i<0x100; i++)
-    if (can_revector_int21(i)==REVECT)
-      set_revectored(i, &vm86s.int21_revectored);
-  saved_int21_revectored = vm86s.int21_revectored;
-  set_int21_revectored(-1);
+  set_int21_revectored(0);
 #endif
 
 #ifdef __NetBSD__
