@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -10,7 +10,7 @@
  * Description: CMOS handling routines
  * 
  * Originally by Robert Sanders, gt8134b@prism.gatech.edu
- * New CMOS code by vignani@mbox.vol.it 1997-98
+ * New CMOS code by vignani@mail.tin.it 1997-98
  *
  * SIDOC_END_MODULE
  *
@@ -60,6 +60,7 @@ cmos_read(ioport_t port)
   case CMOS_DOM:		/* day of month */
   case CMOS_MONTH:
   case CMOS_YEAR:
+  case CMOS_CENTURY:
     holder = cmos_date(cmos.address); goto quit;
 
   case CMOS_CHKSUML:
