@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -41,8 +41,8 @@ struct p_fops def_pfops =
 struct printer lpt[NUM_PRINTERS] =
 {
   {NULL, "lpr", "%s", 5, 0x378},
-  {"lpt2", "lpr", "%s", 5, 0x278},
-  {"lpt3", NULL, NULL, 10, 0x3bc}
+  {NULL, "lpr", "%s", 5, 0x278},
+  {NULL, NULL, NULL, 10, 0x3bc}
 };
 
 void
@@ -92,7 +92,7 @@ printer_open(int prnum)
   um = umask(026);
   if (lpt[prnum].file == NULL) {
     if (!lpt[prnum].dev) {
-      lpt[prnum].dev = tmpnam(NULL);
+      lpt[prnum].dev = strdup(tmpnam(NULL));
       enter_priv_off();
       lpt[prnum].file = fopen(lpt[prnum].dev, "a");
       leave_priv_setting();
@@ -123,6 +123,7 @@ printer_close(int prnum)
   /* delete any temporary files */
   if (lpt[prnum].prtcmd && lpt[prnum].dev) {
     unlink(lpt[prnum].dev);
+    free(lpt[prnum].dev);
     lpt[prnum].dev = NULL;
   }
 
