@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -9,7 +9,7 @@
  *
  * Description: CMOS handling and RTC emulation
  * 
- * Maintainers: Alberto Vignani (vignani@mbox.vol.it)
+ * Maintainers: Alberto Vignani (vignani@mail.tin.it)
  *
  * SIDOC_END_MODULE
  *
@@ -76,6 +76,9 @@ int cmos_date(int reg)
     else
       return BCD(tm->tm_year);
 
+  case CMOS_CENTURY:
+    return BCD(tm->tm_year/100 + 19);
+
   default:
     h_printf("CMOS: cmos_time() register 0x%02x defaulted to 0\n", reg);
     return 0;
@@ -176,7 +179,7 @@ static struct tm *sys_time_calib (void)
 {
   struct tm *tm;
   time_t this_time;
-  unsigned long long k;
+  unsigned long long k=0;	/* gcc warning? */
 
   time(&this_time);
   tm = localtime(&this_time);
