@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -25,6 +25,7 @@
 #include "bios.h"
 #include "kversion.h"
 #include "lpt.h"
+#include "int.h"
 
 #include "dos2linux.h"
 #include "priv.h"
@@ -124,12 +125,44 @@ config_defaults(void)
         config.realcpu = CPU_586;
         config.pci = 1;	/* fair guess */
         cpuflags = get_proc_string_by_key("flags");
-        if (strstr(cpuflags, "tsc")) {
+        if (cpuflags && strstr(cpuflags, "tsc")) {
           /* bogospeed currently returns 0; should it deny
            * pentium features, fall back into 486 case */
-          if (!bogospeed(&config.cpu_spd, &config.cpu_tick_spd)) {
-              config.rdtsc = 1;
+#if LX_KERNEL_VERSION > 2001126
+	  if ((kernel_version_code > 0x20100+126)
+	       && (cpuflags = get_proc_string_by_key("cpu MHz"))) {
+	    int di,df;
+	    if (sscanf(cpuflags,"%d.%d",&di,&df)==2) {
+		/* speed division factor to get 1us from CPU clocks - for
+		 * details on fast division see timers.h */
+		long long chz = 0;		/* strange gcc warning */
+		chz = (di * 1000000) + df;
+		config.cpu_spd = (LLF_US*1000000)/chz;
+
+		/* speed division factor to get 838ns from CPU clocks */
+		config.cpu_tick_spd = (LLF_TICKS*1000000)/chz;
+
+		fprintf (stderr,"kernel CPU speed is %Ld Hz\n",chz);
+/*		fprintf (stderr,"CPU speed factors %ld,%ld\n",
+			config.cpu_spd, config.cpu_tick_spd); */
+#ifdef X86_EMULATOR
+		config.emuspeed = di;
+		fprintf (stderr,"CPU-EMU speed is %d MHz\n",di);
+#endif
+		break;
+	    }
+	    else
+	      cpuflags=NULL;
+	  }
+	  else
+	    cpuflags=0;
+	  if (!cpuflags) {
+#else
+	  if (1) {
+#endif
+            if (!bogospeed(&config.cpu_spd, &config.cpu_tick_spd)) {
               break;
+            }
           }
         }
         /* fall thru */
@@ -146,13 +179,11 @@ config_defaults(void)
       advance_proc_bufferptr();
     }
     if (k > 1) {
-      fprintf(stderr,"Denying use of pentium timer on SMP machine\n");
       config.smp = 1;		/* for checking overrides, later */
-      config.rdtsc = 0;
     }
     close_proc_scan();
-    fprintf(stderr,"Running on CPU=%d86, FPU=%d, rdtsc=%d\n",
-      config.realcpu, config.mathco, config.rdtsc);
+    fprintf(stderr,"Running on CPU=%d86, FPU=%d\n",
+      config.realcpu, config.mathco);
 
     config.hdiskboot = 1;	/* default hard disk boot */
 #ifdef X86_EMULATOR
@@ -253,6 +284,7 @@ config_defaults(void)
     config.emuini = (char *) NULL;
     config.dosbanner = 1;
     config.allowvideoportaccess = 0;
+    config.emuretrace = 0;
 
     config.keytable = &keytable_list[KEYB_USER]; /* What's the current keyboard  */
 
@@ -260,7 +292,8 @@ config_defaults(void)
 				 * new VT. */
     config.debugout = NULL;	/* default to no debug output file */
 
-    config.pre_stroke =NULL;	/* defualt no keyboard pre-strokes */
+    config.pre_stroke =NULL;	/* default no keyboard pre-strokes */
+    config.pre_stroke_mem =NULL;
 
     config.sillyint = 0;
     config.must_spare_hardware_ram = 0;
@@ -290,57 +323,70 @@ config_defaults(void)
     config.mpu401_base = 0x330;
 
     config.vnet = 1;
+
+    memset(config.features, 0, sizeof(config.features));
 }
 
-void dump_config_status(void)
+static void dump_printf(char *fmt, ...)
+{
+    va_list args;
+    va_start(args, fmt);
+    vfprintf(stderr, fmt, args);
+    va_end(args);
+}
+
+void dump_config_status(void *printfunc)
 {
     char *s;
-    FILE *out = stderr;
+    void (*print)(char *, ...) = (printfunc ? printfunc : dump_printf);
 
-    fprintf(out, "\n-------------------------------------------------------------\n");
-    fprintf(out, "------dumping the runtime configuration _after_ parsing -----\n");
-    fprintf(out, "Version: dosemu-" VERSTR " versioncode = 0x%08x\n\n", DOSEMU_VERSION_CODE);
-    fprintf(out, "cpu ");
+    if (!printfunc) {
+      (*print)("\n-------------------------------------------------------------\n");
+      (*print)("------dumping the runtime configuration _after_ parsing -----\n");
+    }
+    else (*print)("dumping the current runtime configuration:\n");
+    (*print)("Version: dosemu-" VERSTR " versioncode = 0x%08x\n\n", DOSEMU_VERSION_CODE);
+    (*print)("cpu ");
     switch (vm86s.cpu_type) {
       case CPU_386: s = "386"; break;
       case CPU_486: s = "486"; break;
       case CPU_586: s = "586"; break;
       default:  s = "386"; break;
     }
-    fprintf(out, "%s\nrealcpu ", s);
+    (*print)("%s\nrealcpu ", s);
     switch (config.realcpu) {
       case CPU_386: s = "386"; break;
       case CPU_486: s = "486"; break;
       case CPU_586: s = "586"; break;
       default:  s = "386"; break;
     }
-    fprintf(out, "%s\n", s);
-    fprintf(out, "CPUclock %gMHz\ncpu_spd 0x%lx\ncpu_tick_spd 0x%lx\n",
+    (*print)("%s\n", s);
+    (*print)("CPUclock %gMHz\ncpu_spd 0x%lx\ncpu_tick_spd 0x%lx\n",
 	(((double)LLF_US)/config.cpu_spd), config.cpu_spd, config.cpu_tick_spd);
 
-    fprintf(out, "pci %d\nrdtsc %d\nmathco %d\nsmp %d\n",
+    (*print)("pci %d\nrdtsc %d\nmathco %d\nsmp %d\n",
                  config.pci, config.rdtsc, config.mathco, config.smp);
 #ifdef X86_EMULATOR
-    fprintf(out, "emuspeed %d\ncpuemu %d\n", config.emuspeed, config.cpuemu);
+    (*print)("emuspeed %d\ncpuemu %d\n", config.emuspeed, config.cpuemu);
 #endif
 
-    fprintf(out, "hdiskboot %d\nmem_size %d\n",
+    (*print)("hdiskboot %d\nmem_size %d\n",
         config.hdiskboot, config.mem_size);
-    fprintf(out, "ems_size 0x%x\nems_frame 0x%x\nsecure %d\n",
+    (*print)("ems_size 0x%x\nems_frame 0x%x\nsecure %d\n",
         config.ems_size, config.ems_frame, config.secure);
-    fprintf(out, "xms_size 0x%x\nmax_umb 0x%x\ndpmi 0x%x\n",
+    (*print)("xms_size 0x%x\nmax_umb 0x%x\ndpmi 0x%x\n",
         config.xms_size, config.max_umb, config.dpmi);
-    fprintf(out, "mouse_flag %d\nmapped_bios %d\nvbios_file %s\n",
+    (*print)("mouse_flag %d\nmapped_bios %d\nvbios_file %s\n",
         config.mouse_flag, config.mapped_bios, (config.vbios_file ? config.vbios_file :""));
-    fprintf(out, "vbios_copy %d\nvbios_seg 0x%x\nvbios_size 0x%x\n",
+    (*print)("vbios_copy %d\nvbios_seg 0x%x\nvbios_size 0x%x\n",
         config.vbios_copy, config.vbios_seg, config.vbios_size);
-    fprintf(out, "console %d\nconsole_keyb %d\nconsole_video %d\n",
+    (*print)("console %d\nconsole_keyb %d\nconsole_video %d\n",
         config.console, config.console_keyb, config.console_video);
-    fprintf(out, "kbd_tty %d\nexitearly %d\n",
+    (*print)("kbd_tty %d\nexitearly %d\n",
         config.kbd_tty, config.exitearly);
-    fprintf(out, "fdisks %d\nhdisks %d\nbootdisk %d\n",
+    (*print)("fdisks %d\nhdisks %d\nbootdisk %d\n",
         config.fdisks, config.hdisks, config.bootdisk);
-    fprintf(out, "term_esc_char 0x%x\nterm_color %d\nterm_updatefreq %d\n",
+    (*print)("term_esc_char 0x%x\nterm_color %d\nterm_updatefreq %d\n",
         config.term_esc_char, config.term_color, config.term_updatefreq);
     switch (config.term_charset) {
       case CHARSET_LATIN: s = "latin"; break;
@@ -350,23 +396,23 @@ void dump_config_status(void)
       case CHARSET_FULLIBM: s = "fullibm"; break;
       default: s = "(unknown)";
     }
-    fprintf(out, "term_charset \"%s\"\nX_updatelines %d\nX_updatefreq %d\n",
+    (*print)("term_charset \"%s\"\nX_updatelines %d\nX_updatefreq %d\n",
         s, config.X_updatelines, config.X_updatefreq);
-    fprintf(out, "X_display \"%s\"\nX_title \"%s\"\nX_icon_name \"%s\"\n",
+    (*print)("X_display \"%s\"\nX_title \"%s\"\nX_icon_name \"%s\"\n",
         (config.X_display ? config.X_display :""), config.X_title, config.X_icon_name);
-    fprintf(out, "X_blinkrate %d\nX_sharecmap %d\nX_mitshm %d\n",
+    (*print)("X_blinkrate %d\nX_sharecmap %d\nX_mitshm %d\n",
         config.X_blinkrate, config.X_sharecmap, config.X_mitshm);
-    fprintf(out, "X_fixed_aspect %d\nX_aspect_43 %d\nX_lin_filt %d\n",
+    (*print)("X_fixed_aspect %d\nX_aspect_43 %d\nX_lin_filt %d\n",
         config.X_fixed_aspect, config.X_aspect_43, config.X_lin_filt);
-    fprintf(out, "X_bilin_filt %d\nX_mode13fact %d\nX_winsize_x %d\n",
+    (*print)("X_bilin_filt %d\nX_mode13fact %d\nX_winsize_x %d\n",
         config.X_bilin_filt, config.X_mode13fact, config.X_winsize_x);
-    fprintf(out, "X_winsize_y %d\nX_gamma %d\nvgaemu_memsize 0x%x\n",
+    (*print)("X_winsize_y %d\nX_gamma %d\nvgaemu_memsize 0x%x\n",
         config.X_winsize_y, config.X_gamma, config.vgaemu_memsize);
-    fprintf(out, "vesamode_list %p\nX_lfb %d\nX_pm_interface %d\n",
+    (*print)("vesamode_list %p\nX_lfb %d\nX_pm_interface %d\n",
         config.vesamode_list, config.X_lfb, config.X_pm_interface);
-    fprintf(out, "X_keycode %d\nX_font \"%s\"\nusesX %d\n",
+    (*print)("X_keycode %d\nX_font \"%s\"\nusesX %d\n",
         config.X_keycode, config.X_font, config.usesX);
-    fprintf(out, "X_mgrab_key \"%s\"\n",  config.X_mgrab_key);
+    (*print)("X_mgrab_key \"%s\"\n",  config.X_mgrab_key);
     switch (config.chipset) {
       case PLAINVGA: s = "plainvga"; break;
       case TRIDENT: s = "trident"; break;
@@ -380,7 +426,7 @@ void dump_config_status(void)
       case WDVGA: s = "wdvga"; break;
       default: s = "unknown"; break;
     }
-    fprintf(out, "config.X %d\nhogthreshold %d\nchipset \"%s\"\n",
+    (*print)("config.X %d\nhogthreshold %d\nchipset \"%s\"\n",
         config.X, config.hogthreshold, s);
     switch (config.cardtype) {
       case CARD_VGA: s = "VGA"; break;
@@ -389,9 +435,9 @@ void dump_config_status(void)
       case CARD_EGA: s = "EGA"; break;
       default: s = "unknown"; break;
     }
-    fprintf(out, "cardtype \"%s\"\npci_video %d\nfullrestore %d\n",
+    (*print)("cardtype \"%s\"\npci_video %d\nfullrestore %d\n",
         s, config.pci_video, config.fullrestore);
-    fprintf(out, "graphics %d\ngfxmemsize %d\nvga %d\n",
+    (*print)("graphics %d\ngfxmemsize %d\nvga %d\n",
         config.graphics, config.gfxmemsize, config.vga);
     switch (config.speaker) {
       case SPKR_OFF: s = "off"; break;
@@ -399,48 +445,48 @@ void dump_config_status(void)
       case SPKR_EMULATED: s = "emulated"; break;
       default: s = "wrong"; break;
     }
-    fprintf(out, "dualmon %d\nforce_vt_switch %d\nspeaker \"%s\"\n",
+    (*print)("dualmon %d\nforce_vt_switch %d\nspeaker \"%s\"\n",
         config.dualmon, config.force_vt_switch, s);
-    fprintf(out, "update %d\nfreq %d\nwantdelta %d\nrealdelta %d\n",
+    (*print)("update %d\nfreq %d\nwantdelta %d\nrealdelta %d\n",
         config.update, config.freq, config.wantdelta, config.realdelta);
-    fprintf(out, "timers %d\nkeybint %d\n",
+    (*print)("timers %d\nkeybint %d\n",
         config.timers, config.keybint);
-    fprintf(out, "tty_lockdir \"%s\"\ntty_lockfile \"%s\"\nconfig.tty_lockbinary %d\n",
+    (*print)("tty_lockdir \"%s\"\ntty_lockfile \"%s\"\nconfig.tty_lockbinary %d\n",
         config.tty_lockdir, config.tty_lockfile, config.tty_lockbinary);
-    fprintf(out, "num_ser %d\nnum_lpt %d\nnum_mice %d\nfastfloppy %d\n",
+    (*print)("num_ser %d\nnum_lpt %d\nnum_mice %d\nfastfloppy %d\n",
         config.num_ser, config.num_lpt, config.num_mice, config.fastfloppy);
-    fprintf(out, "emusys \"%s\"\nemubat \"%s\"\nemuini \"%s\"\n",
+    (*print)("emusys \"%s\"\nemubat \"%s\"\nemuini \"%s\"\n",
         (config.emusys ? config.emusys : ""), (config.emubat ? config.emubat : ""), (config.emuini ? config.emuini : ""));
-    fprintf(out, "dosbanner %d\nallowvideoportaccess %d\ndetach %d\n",
+    (*print)("dosbanner %d\nallowvideoportaccess %d\ndetach %d\n",
         config.dosbanner, config.allowvideoportaccess, config.detach);
-    fprintf(out, "debugout \"%s\"\n",
+    (*print)("debugout \"%s\"\n",
         (config.debugout ? config.debugout : (unsigned char *)""));
     {
 	char buf[256];
 	GetDebugFlagsHelper(buf, 0);
-	fprintf(out, "debug_flags \"%s\"\n", buf);
+	(*print)("debug_flags \"%s\"\n", buf);
     }
     {
       extern void dump_keytable(FILE *f, struct keytable_entry *kt);
-      dump_keytable(out, config.keytable);
+      if (!printfunc) dump_keytable(stderr, config.keytable);
     }
-    fprintf(out, "pre_stroke \"%s\"\n", (config.pre_stroke ? config.pre_stroke : (unsigned char *)""));
-    fprintf(out, "irqpassing= ");
+    (*print)("pre_stroke \"%s\"\n", (config.pre_stroke ? config.pre_stroke : (unsigned char *)""));
+    (*print)("irqpassing= ");
     if (config.sillyint) {
       int i;
       for (i=0; i <16; i++) {
         if (config.sillyint & (1<<i)) {
-          fprintf(out, "IRQ%d", i);
+          (*print)("IRQ%d", i);
           if (config.sillyint & (0x10000<<i))
-            fprintf(out, "(sigio) ");
+            (*print)("(sigio) ");
           else
-            fprintf(out, " ");
+            (*print)(" ");
         }
       }
-      fprintf(out, "\n");
+      (*print)("\n");
     }
-    else fprintf(out, "none\n");
-    fprintf(out, "must_spare_hardware_ram %d\n",
+    else (*print)("none\n");
+    (*print)("must_spare_hardware_ram %d\n",
         config.must_spare_hardware_ram);
     {
       int need_header_line =1;
@@ -448,15 +494,15 @@ void dump_config_status(void)
       for (i=0; i<(sizeof(config.hardware_pages)); i++) {
         if (config.hardware_pages[i]) {
           if (need_header_line) {
-            fprintf(out, "hardware_pages:\n");
+            (*print)("hardware_pages:\n");
             need_header_line = 0;
           }
-          fprintf(out, "%05x ", (i << 12) + 0xc8000);
+          (*print)("%05x ", (i << 12) + 0xc8000);
         }
       }
-      if (!need_header_line) fprintf(out, "\n");
+      if (!need_header_line) (*print)("\n");
     }
-    fprintf(out, "ipxsup %d\nvnet %d\npktflags 0x%x\n",
+    (*print)("ipxsup %d\nvnet %d\npktflags 0x%x\n",
 	config.ipxsup, config.vnet, config.pktflags);
     
     {
@@ -464,15 +510,25 @@ void dump_config_status(void)
 	struct printer *pptr;
 	extern struct printer lpt[NUM_PRINTERS];
 	for (i = 0, pptr = lpt; i < config.num_lpt; i++, pptr++) {
-	  fprintf(out, "LPT%d command \"%s  %s\"  timeout %d  device \"%s\"  baseport 0x%03x\n",
+	  (*print)("LPT%d command \"%s  %s\"  timeout %d  device \"%s\"  baseport 0x%03x\n",
 	  i+1, pptr->prtcmd, pptr->prtopt, pptr->delay, (pptr->dev ? pptr->dev : ""), pptr->base_port); 
 	}
     }
 
-    fprintf(out, "\nSOUND:\nsb_base 0x%x\nsb_dma %d\nsb_irq %d\nmpu401_base 0x%x\nsb_dsp \"%s\"\nsb_mixer \"%s\"\n",
+    {
+	int i;
+	int size = sizeof(config.features) / sizeof(config.features[0]);
+	for (i = 0; i < size; i++) {
+	  (*print)("feature_%d %d\n", i, config.features[i]);
+	}
+    }
+
+    (*print)("\nSOUND:\nsb_base 0x%x\nsb_dma %d\nsb_irq %d\nmpu401_base 0x%x\nsb_dsp \"%s\"\nsb_mixer \"%s\"\n",
         config.sb_base, config.sb_dma, config.sb_irq, config.mpu401_base, config.sb_dsp, config.sb_mixer);
-    fprintf(out, "\n--------------end of runtime configuration dump -------------\n");
-    fprintf(out,   "-------------------------------------------------------------\n\n");
+    if (!printfunc) {
+      (*print)("\n--------------end of runtime configuration dump -------------\n");
+      (*print)(  "-------------------------------------------------------------\n\n");
+    }
 }
 
 static void 
@@ -541,6 +597,9 @@ static void our_envs_init(char *usedoptions)
     setenv("DOSEMU_EUID", buf, 1);
     sprintf(buf, "%d", getuid());
     setenv("DOSEMU_UID", buf, 1);
+    strcpy(buf, "0");
+    if (is_console(0)) strcpy(buf, "1");
+    setenv("DOSEMU_STDIN_IS_CONSOLE", buf, 1);
 }
 
 
@@ -569,7 +628,7 @@ config_init(int argc, char **argv)
     PRIV_SAVE_AREA
     extern char *commandline_statements;
     extern int dexe_running;
-    int             c;
+    int             c=0;
     char           *confname = NULL;
     char           *dosrcname = NULL;
     char           *basename;
@@ -606,7 +665,7 @@ config_init(int argc, char **argv)
 
     opterr = 0;
     confname = CONFIG_SCRIPT;
-    while ((c = getopt(argc, argv, "ABCcF:f:I:kM:D:P:VNtsgh:H:x:KL:m23456e:E:dXY:Z:o:Ou:")) != EOF) {
+    while ((c = getopt(argc, argv, "ABCcF:f:I:kM:D:P:VNtsgh:H:x:KL:m23456e:E:dXY:Z:o:Ou:U:")) != EOF) {
 	usedoptions[(unsigned char)c] = c;
 	switch (c) {
 	case 'h':
@@ -700,6 +759,11 @@ config_init(int argc, char **argv)
 		define_config_variable(s);
 	    }
 	    break;
+	case 'U': {
+		extern void init_uhook(char *pipes);
+		init_uhook(optarg);
+	    }
+	    break;
 	}
     }
 
@@ -743,7 +807,9 @@ config_init(int argc, char **argv)
     optind = 0;
 #endif
     opterr = 0;
-    while ((c = getopt(argc, argv, "ABCcF:f:I:kM:D:P:v:VNtsgh:H:x:KLm23456e:dXY:Z:E:o:Ou:")) != EOF) {
+    while ((c = getopt(argc, argv, "ABCcF:f:I:kM:D:P:v:VNtsgh:H:x:KLm23456e:dXY:Z:E:o:Ou:U:")) != EOF) {
+	/* currently _NOT_ used option characters: abGjJlpqQrRSTwWyz */
+
 	/* Note: /etc/dosemu.conf may have disallowed some options
 	 *	 ( by removing them from $DOSEMU_OPTIONS ).
 	 *	 We skip them by re-checking 'usedoptions'.
@@ -762,6 +828,7 @@ config_init(int argc, char **argv)
 	case 'O':
 	case 'L':
 	case 'u':
+	case 'U':
 	case '2': case '3': case '4': case '5': case '6':
 	    break;
 	case 'A':
@@ -896,6 +963,7 @@ config_init(int argc, char **argv)
 	    c_printf("CONF: Forceing neutral Keyboard-layout, X-server will translate\n");
 	}
 	config.console_video = config.vga = config.graphics = 0;
+	config.emuretrace = 0;	/* already emulated */
     }
     else {
 	if (!can_do_root_stuff && config.console) {
@@ -919,8 +987,14 @@ config_init(int argc, char **argv)
     c_printf(" uid=%d (cached %d) gid=%d (cached %d)\n",
         geteuid(), get_cur_euid(), getegid(), get_cur_egid());
 
+#ifdef X86_EMULATOR
+    if (config.cpuemu && config.speaker==SPKR_NATIVE) {
+	c_printf("SPEAKER: can`t use native mode with cpu-emu\n");
+	config.speaker=SPKR_EMULATED;
+    }
+#endif
     if (config_check_only) {
-	dump_config_status();
+	dump_config_status(0);
 	usage();
 	leavedos(0);
     }
@@ -979,9 +1053,9 @@ int parse_debugflags(const char *s, unsigned char flag)
     char            c;
     int ret = 0;
 #ifdef X_SUPPORT
-    const char      allopts[] = "dRWDCvXkiTsm#pgcwhIExMnPrSe";
+    const char      allopts[] = "dARWDCvXkiTsm#pgcwhIExMnPrSe";
 #else
-    const char      allopts[] = "dRWDCvkiTsm#pgcwhIExMnPrSe";
+    const char      allopts[] = "dARWDCvkiTsm#pgcwhIExMnPrSe";
 #endif
 
     /*
@@ -1012,8 +1086,8 @@ int parse_debugflags(const char *s, unsigned char flag)
 	    break;
 	case 'D':		/* DOS int 21h */
 	    d.dos = flag;
-	    { extern void set_int21_revectored(int);
-	      set_int21_revectored(d.dos != 0);
+	    { static int first = 1;
+	      if(first) { set_int21_revectored(d.dos ? 1 : 0); first = 0; }
 	    }
 	    break;
         case 'C':               /* CDROM */
@@ -1034,8 +1108,11 @@ int parse_debugflags(const char *s, unsigned char flag)
 	    d.io = flag;
 	    break;
 	case 'T':		/* i/o port tracing */
+	{   extern void init_port_traceing(void);
 	    d.io_trace = flag;
+	    if (d.io_trace) init_port_traceing();
 	    break;
+	}
 	case 's':		/* serial */
 	    d.serial = flag;
 	    break;
@@ -1084,6 +1161,9 @@ int parse_debugflags(const char *s, unsigned char flag)
 	case 'S':		/* SOUND */
 	    d.sound = flag;
 	    break;
+	case 'A':		/* ASPI */
+	    d.aspi = flag;
+	    break;
 #ifdef X86_EMULATOR
 	case 'e':		/* cpu-emu */
 	    d.emu = flag;
