@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -566,16 +566,28 @@ static Boolean handle_dosemu_keys(t_keysym key) {
 
    if ((shiftstate&CTRL) && (shiftstate&ALT)) {
       switch(key) {
+#ifdef X86_EMULATOR
+       case KEY_PGUP:
+             k_printf("KBD: Ctrl-Alt-PgUp\n");
+             if (config.cpuemu) {
+               if (d.emu<2) d.emu=4;
+               fflush(dbg_fd);
+             }
+             return 1;
+#else
 #if 0	/* C-A-D is disabled */
        case KEY_DEL:
        case KEY_PGUP:
              k_printf("KBD: Ctrl-Alt-{Del|PgUp}: rebooting dosemu\n");
              dos_ctrl_alt_del();
              return 1;
+#endif
 #endif
        case KEY_PGDN:
              k_printf("KBD: Ctrl-Alt-PgDn: bye bye!\n");
-             leavedos(0);
+             leavedos(123);  /* use this exitcode, so a wrapper script
+				may detect Ctrl-Alt-PgDn and restart
+				dosemu if if wants (pseudo-reboot) */
 
        case KEY_F1 ... KEY_F10:
        case KEY_F11:
