@@ -1,3 +1,9 @@
+/* 
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /*
  * Memory allocation routines for DPMI clients.
  *
@@ -9,7 +15,7 @@
 
 #include <stdio.h>		/* for NULL */
 #include <stdlib.h>
-#include <strings.h>		/* for memcpy */
+#include <string.h>		/* for memcpy */
 #include <sys/mman.h>
 #include <sys/types.h>
 #include <fcntl.h>
