@@ -39,7 +39,8 @@ Cambridge, MA 02139, USA.
 The maintainer of the Willows TWIN Libraries may be reached (Email) 
 at the address twin@willows.com	
 
-changes for use with dosemu-0.67 1997/10/20 vignani@mbox.vol.it
+changes for use with dosemu-0.67 1997/10/20 vignani@tin.it
+changes for use with dosemu-0.99 1998/12/13 vignani@tin.it
 
 *********************************************************************/
 
@@ -47,12 +48,10 @@ changes for use with dosemu-0.67 1997/10/20 vignani@mbox.vol.it
 #define hsw_interp__h
 /* "@(#)hsw_interp.h	1.23 :/users/sccs/src/win/intp32/s.hsw_interp.h 1/23/97 17:25:08"  */
 
-#include "windows.h"
-#include "kerndef.h"
-#include "BinTypes.h"
-#include "DPMI.h"
+#include "internal.h"
+#include <setjmp.h>
 
-#ifndef DOSEMU_TYPESONLY
+#if !defined(DOSEMU) || defined(_IN_INTP32)
 
 #define ADDbfrm		0x00
 #define ADDwfrm		0x01
@@ -428,8 +427,11 @@ changes for use with dosemu-0.67 1997/10/20 vignani@mbox.vol.it
 #endif	/* ] */
 
 #define CARRY_FLAG		0x0001
+/* reserved			0x0002 */
 #define PARITY_FLAG		0x0004
+/* reserved			0x0008 */
 #define AUX_CARRY_FLAG		0x0010
+/* reserved			0x0020 */
 #define ZERO_FLAG		0x0040
 #define SIGN_FLAG		0x0080
 #define TRAP_FLAG		0x0100
@@ -438,29 +440,68 @@ changes for use with dosemu-0.67 1997/10/20 vignani@mbox.vol.it
 #define OVERFLOW_FLAG		0x0800
 #define IOPL_FLAG_MASK		0x3000
 #define NESTED_FLAG		0x4000
-#define BYTE_FL			0x8000
+#define BYTE_FL			0x8000	/* Intel reserved! */
+#define RF_FLAG			0x10000
+#define VM_FLAG			0x20000
+/* AC				0x40000 */
+
+#define CPL	0x3000
+#define IOPL	(env->flags&IOPL_FLAG_MASK)
 
 #define BYTE_OP 0x45
 
-#define IS_AF_SET (BOOL)((int)((interp_var->flags_czsp.byte.byte_op==BYTE_OP)?\
-	(((interp_var->src1.byte.byte & 0xf) + (interp_var->src2.byte.byte & 0xf))\
-	& AUX_CARRY_FLAG):\
-	(((interp_var->src1.word16.word & 0xf) + (interp_var->src2.word16.word & 0xf))\
-	& AUX_CARRY_FLAG)) >> 4)
-#define IS_CF_SET (BOOL)(interp_var->flags_czsp.word16.carry & CARRY_FLAG)
-#define IS_OF_SET ((WORD)((~(interp_var->src1.word16.word ^ interp_var->src2.word16.word))&(interp_var->src1.word16.word ^ interp_var->flags_czsp.word16.res16))>>15)
-/* #define _IS_OF_SET (((interp_var->src1.word16.word>>15)==(interp_var->src2.word16.word>>15))&&((interp_var->src1.word16.word>>15)!=(interp_var->flags_czsp.word16.res16>>15))) */
-#define IS_ZF_SET (BOOL)(interp_var->flags_czsp.word16.res16 == 0)
-#define IS_SF_SET (BOOL)(interp_var->flags_czsp.byte.res8 >> 7)
-#define IS_PF_SET (parity[((interp_var->flags_czsp.byte.byte_op==BYTE_OP)?interp_var->flags_czsp.byte.res8:interp_var->flags_czsp.byte.parity16)]==PARITY_FLAG)
+/*
+ * Test condition uses:	RES	SRC1,SRC2
+ *	Z,NZ		 Y	    N
+ *	S,NS		 Y	    N
+ *	CY,NC		 Y	    N
+ *	P,NP		 Y	    N
+ *	OV,NO		 Y	    Y
+ *	AF		 N	    Y
+ *	CX,LOOP		 N	    N
+ *
+ * beware of logical operations - all conditions _should_ return 0 or 1
+ *
+ *		[                  ][------word--------] SRCn_16
+ *		[        ][        ][--byte--][--aux---]
+ *				      SRCn_8
+ */
+#define IS_AF_SET (((SRC1_16 & 0xf0f) + (SRC2_16 & 0xf0f)) & \
+	(BYTE_FLAG==BYTE_OP? 0x1000:0x10))
 
-#define CLEAR_CF interp_var->flags_czsp.byte.carryb = 0
-#define CLEAR_AF { interp_var->src1.word16.word &= ~0x808; interp_var->src2.word16.word &= ~0x808; }
-#define CLEAR_DF env->flags &= (~ DIRECTION_FLAG)
+#define IS_CF_SET (BOOL)(CARRY & CARRY_FLAG)
 
-#define SET_CF interp_var->flags_czsp.byte.carryb = 1
-#define SET_AF { interp_var->src1.word16.word |= 0x808; interp_var->src2.word16.word |= 0x808; }
+/* overflow rule using bit15:
+ *	src1 src2  res	b15
+ *	  0    0    0    0
+ *	  0    0    1    1	BUT stc;sbb 0,0 -> ..ff, OF=0
+ *	  0    1    0    0
+ *	  0    1    1    0
+ *	  1    0    0    0	BUT stc;sbb 80..,0 -> 7f.., OF=1
+ *	  1    0    1    0
+ *	  1    1    0    1
+ *	  1    1    1    0
+ */ 
+#define IS_OF_SET (BOOL)((((~(SRC1_16 ^ SRC2_16))&(SRC1_16 ^ RES_16)) >> 15) & 1)
+
+#define IS_ZF_SET (BOOL)(RES_16 == 0)
+#define IS_SF_SET (BOOL)((RES_8 >> 7) & 1)
+#define IS_PF_SET (parity[((BYTE_FLAG==BYTE_OP)?RES_8:PAR_16)]==PARITY_FLAG)
+
+#define CLEAR_CF CARRYB=0
+/* Wow. this changes also parity */
+#define CLEAR_AF { SRC1_16 &= ~0x808; SRC2_16 &= ~0x808; }
+#define CLEAR_DF env->flags &= ~DIRECTION_FLAG
+/* this is safe */
+#define CLEAR_ZF RES_8|=1
+
+#define SET_CF CARRYB=1
+/* Wow. this changes also parity */
+#define SET_AF { SRC1_16 |= 0x808; SRC2_16 |= 0x808; }
 #define SET_DF env->flags |= DIRECTION_FLAG
+/* this way of setting ZF affects also SF,OF,PF,AF. Some operations like
+ * LAR,LSL specify that ONLY ZF is affected. God save the Queen. */
+#define SET_ZF RES_16=0
 
 #define ROL	0
 #define ROR	1
@@ -532,14 +573,21 @@ changes for use with dosemu-0.67 1997/10/20 vignani@mbox.vol.it
 #define SHORT_FS_32 env->fs.fs
 #define SHORT_GS_16 env->gs.word16.gs
 #define SHORT_GS_32 env->gs.gs
-#define LONG_CS interp_var->seg_regs[0]
-#define LONG_DS interp_var->seg_regs[1]
-#define LONG_ES interp_var->seg_regs[2]
-#define LONG_SS interp_var->seg_regs[3]
-#define LONG_FS interp_var->seg_regs[4]
-#define LONG_GS interp_var->seg_regs[5]
+#define LONG_CS env->seg_regs[0]
+#define LONG_DS env->seg_regs[1]
+#define LONG_ES env->seg_regs[2]
+#define LONG_SS env->seg_regs[3]
+#define LONG_FS env->seg_regs[4]
+#define LONG_GS env->seg_regs[5]
+#define BIG_CS env->bigseg[0]
+#define BIG_DS env->bigseg[1]
+#define BIG_ES env->bigseg[2]
+#define BIG_SS env->bigseg[3]
+#define BIG_FS env->bigseg[4]
+#define BIG_GS env->bigseg[5]
 
-#define OVERRIDE interp_var->seg_regs[6]
+#define OVERR_DS env->seg_regs[6]
+#define OVERR_SS env->seg_regs[7]
 #define INVALID_OVR	((unsigned char *)(-1l))
 
 #define SIGNb (res & 0x80)
@@ -561,7 +609,9 @@ changes for use with dosemu-0.67 1997/10/20 vignani@mbox.vol.it
 #define AT_READ		2
 #define AT_WRITE	3
 
+#ifndef DOSEMU
 #define ERRORLOG(s,o,m,r) fprintf(stderr,"Error accessing selector %x, offset %x, mode %x, errorcode %x\n",s,o,m,r)
+#endif
 
 #define UINT_86		2
 #define INT_86		2
@@ -569,88 +619,129 @@ changes for use with dosemu-0.67 1997/10/20 vignani@mbox.vol.it
 
 #define JUMP(lp)	PC = PC + 2 +(signed char)(*(lp)); 
 
-#if defined(DOSEMU) && defined(__i386__)
-#define PUSHWORD(w) {env->rsp.sp.sp-=2; \
-		*((unsigned short *)(LONG_SS+env->rsp.sp.sp))=(w);}
-#define PUSHQUAD(dw) {env->rsp.sp.sp-=4; \
-		*((unsigned long *)(LONG_SS+env->rsp.sp.sp))=(dw);}
+#ifdef __i386__
+#define TOS_WORD	(BIG_SS? *((unsigned short *)(LONG_SS+ESP)):\
+			*((unsigned short *)(LONG_SS+SP)))
 
-#define POPWORD(w)  {w=*((unsigned short *)(LONG_SS+env->rsp.sp.sp));\
-		env->rsp.sp.sp+=2;}
-#define POPQUAD(dw)  {dw=*((unsigned long *)(LONG_SS+env->rsp.sp.sp));\
-		env->rsp.sp.sp+=4;}
-#else
-#define PUSHWORD(w) {unsigned char *sp=interp_var->seg_regs[3]+env->rsp.sp.sp; \
+#define PUSHWORD(w) if (BIG_SS) {ESP-=2; \
+		*((unsigned short *)(LONG_SS+ESP))=(unsigned short)(w);} else \
+		{SP-=2; *((unsigned short *)(LONG_SS+SP))=(unsigned short)(w);}
+#define PUSHQUAD(dw) if (BIG_SS) {ESP-=4; \
+		*((unsigned long *)(LONG_SS+ESP))=(dw);} else \
+		{SP-=4; *((unsigned long *)(LONG_SS+SP))=(dw);}
+
+#define POPWORD(w)  if (BIG_SS) {w=*((unsigned short *)(LONG_SS+ESP));\
+		ESP+=2;} else {w=*((unsigned short *)(LONG_SS+SP)); SP+=2;}
+#define POPQUAD(dw)  if (BIG_SS) {dw=*((unsigned long *)(LONG_SS+ESP));\
+		ESP+=4;} else {dw=*((unsigned long *)(LONG_SS+SP)); SP+=4;}
+#else	/* not i386 */
+#define PUSHWORD(w) {unsigned char *sp=env->seg_regs[3]+env->rsp.sp.sp; \
 		*(sp-2)=w; \
 		*(sp-1)=w>>8; \
 		env->rsp.sp.sp-=2;}
-#define PUSHQUAD(w) {unsigned char *sp=interp_var->seg_regs[3]+env->rsp.sp.sp; \
+#define PUSHQUAD(w) {unsigned char *sp=env->seg_regs[3]+env->rsp.sp.sp; \
 		*(sp-4)=w; \
 		*(sp-3)=w>>8; \
 		*(sp-2)=w>>16; \
 		*(sp-1)=w>>24; \
 		env->rsp.sp.sp-=4;}
 
-#define POPWORD(w) {unsigned char *sp=interp_var->seg_regs[3]+env->rsp.sp.sp; \
+#define POPWORD(w) {unsigned char *sp=env->seg_regs[3]+env->rsp.sp.sp; \
 		w = *(sp) | (((unsigned short) (*(sp+1)))<<8); \
 		env->rsp.sp.sp+=2;}
-#define POPQUAD(w) {unsigned char *sp=interp_var->seg_regs[3]+env->rsp.sp.sp; \
+#define POPQUAD(w) {unsigned char *sp=env->seg_regs[3]+env->rsp.sp.sp; \
 		w = *(sp) | (((unsigned short) *(sp+1))<<8) | \
 			(((unsigned long) *(sp+2))<<16) | \
 			(((unsigned long) *(sp+3))<<24); \
 		env->rsp.sp.sp+=4;}
-#endif	/* DOSEMU */
+#endif
 
 #define SELECTOR_PADDRESS(sel) GetPhysicalAddress(sel)
 
-#ifdef DOSEMU
-#define SET_SEGREG(lp,sel) { \
-	WORD wFlags; \
-	if (vm86f) lp = (unsigned char *)(sel<<4); \
-	else if ((sel >> 3) == 0) \
-	    lp = 0; \
-	else { \
-	    wFlags = GetSelectorFlags(sel); \
-	    if (!(wFlags & DF_PRESENT)) { \
-	        if (!LoadSegment(sel)) { \
-		    error("INTP32: failed to load selector %x\n",(int)sel); \
-		    FatalAppExit(0,NULL); \
-	        } \
-	    } \
-	    lp = GetPhysicalAddress(sel); \
-	} \
-}
-#else
-#define SET_SEGREG(lp,sel) { \
-	WORD wFlags; \
-	char prnt_buf[40]; \
-	if ((sel >> 3) == 0) \
-	    lp = 0; \
-	else { \
-	    wFlags = GetSelectorFlags(sel); \
-	    if (!(wFlags & DF_PRESENT)) { \
-	        if (!LoadSegment(sel)) { \
-		    sprintf(prnt_buf, \
-		        "INTP32: failed to load selector %x\n",sel); \
-		    printf("%s\n",prnt_buf); \
-		    FatalAppExit(0,prnt_buf); \
-	        } \
-	    } \
-	    lp = GetPhysicalAddress(sel); \
-	} \
-}
-#endif	/* DOSEMU */
+extern int SetSegreg(unsigned char **lp, unsigned char *big,
+	unsigned long csel);
+#define SET_SEGREG(lp,big,mk,sel)	SetSegreg(&(lp),&(big),(mk|sel))
 
-#endif	/* DOSEMU_TYPESONLY */
+#ifdef DOSEMU
+int hsw_verr(unsigned short sel);
+int hsw_verw(unsigned short sel);
+#else
+#define hsw_verr(s)	(1)
+#define hsw_verw(s)	(1)
+#endif
+
+#endif
+
+/* ======================= target-specific stuff ====================== */
+
+#ifdef __i386__
+/* use true 80-bit registers on i386 */
+typedef	long double	Ldouble;
+#else
+/* use 64-bit doubles on other targets */
+typedef	double	Ldouble;
+#endif
 
 typedef struct tagENV87
-{   double          fpregs[8];  /* floating point register stack */
+{   	Ldouble         fpregs[8];  /* floating point register stack */
 	int             fpstt;      /* top of register stack */
-	unsigned short  fpus, fpuc;
+	unsigned short  fpus, fpuc, fptag;
 } ENV87;
 
 
-#if defined(sparc) || defined(mips) || defined(ppc) || defined(hppa)
+/* X386 layout:
+ *		[-----------------res------------------] RES_32
+ *	CARRY	[------carry-------][------res16-------] RES_16
+ *		[byte_op-][-carryb-][--res8--][parity16]
+ *		BYTE_FLAG	      RES_8
+ */
+#if defined(sparc) || defined(mips) || defined(ppc) || defined(hppa) || defined(arm)
+typedef	union {
+		unsigned long res;
+		struct {
+			unsigned short carry;
+			unsigned short res16;
+		} word16;
+		struct {
+			unsigned char byte_op;
+			unsigned char carryb;
+			unsigned char res8;
+			unsigned char parity16;
+		} byte;
+	} Interp_VAR_flags_czsp;
+#elif defined(alpha)
+typedef	union {
+		unsigned long res;
+		struct {
+			unsigned short res16;
+			unsigned short carry;
+			unsigned int dummy1;
+		} word16;
+		struct {
+			unsigned char parity16;
+			unsigned char res8;
+			unsigned char carryb;
+			unsigned char byte_op;
+			unsigned int dummy1;
+		} byte;
+	} Interp_VAR_flags_czsp;
+#elif defined(__i386__) 
+typedef	union {
+		unsigned long res;
+		struct {
+			unsigned short res16;
+			unsigned short carry;
+		} word16;
+		struct {
+			unsigned char parity16;
+			unsigned char res8;
+			unsigned char carryb;
+			unsigned char byte_op;
+		} byte;
+	} Interp_VAR_flags_czsp;
+#endif
+
+#if defined(sparc) || defined(mips) || defined(ppc) || defined(hppa) || defined(arm)
 typedef struct keyENV{
 	union {
 		unsigned int ds;
@@ -784,10 +875,41 @@ typedef struct keyENV{
 	unsigned long   level;
 	struct HSW_86_CATCHBUF *buf;
 	jmp_buf		jump_buffer;
+	unsigned char *seg_regs[8];
+	Interp_VAR_flags_czsp flags_czsp;
+	union {
+		unsigned long longword;
+		struct {
+			unsigned short dummy;
+			unsigned short word;
+		} word16;
+		struct {
+			unsigned char dummy1;
+			unsigned char dummy2;
+			unsigned char byte;
+ 			unsigned char aux;
+		} byte;
+	} src1;
+	union {
+		unsigned long longword;
+		struct {
+			unsigned short dummy;
+			unsigned short word;
+		} word16;
+		struct {
+			unsigned char dummy1;
+			unsigned char dummy2;
+			unsigned char byte;
+ 			unsigned char aux;
+		} byte;
+	} src2;
+	unsigned char *reg1;	/* from REG field of mod_rm */
+	unsigned char *mem_ref;	/* from RM field of mod_rm */
+	int mode_reg;
 } Interp_ENV;
 #endif
 
-#if defined(X386) 
+#if defined(__i386__) 
 typedef struct keyENV{
 	union {
 		unsigned int ds;
@@ -914,6 +1036,7 @@ typedef struct keyENV{
 	} gs;
 	BINADDR		trans_addr;
 	BINADDR		return_addr;
+#ifndef DOSEMU
 	LPVOID		machine_stack;
 	struct ENV	*prev_env;
 	unsigned long   active;
@@ -921,6 +1044,52 @@ typedef struct keyENV{
 	unsigned long   level;
 	struct HSW_86_CATCHBUF *buf;
 	jmp_buf		jump_buffer;
+#endif
+	unsigned char *seg_regs[8];
+	Interp_VAR_flags_czsp flags_czsp;
+/*
+ *		[--------------longword----------------] SRC1_32
+ *		[                  ][------word--------] SRC1_16
+ *		[        ][        ][--byte--][--aux---]
+ *				      SRC1_8    AUX1_8
+ */
+	union {
+		unsigned long longword;
+		struct {
+			unsigned short word;
+			unsigned short dummy;
+		} word16;
+		struct {
+ 			unsigned char aux;
+			unsigned char byte;
+			unsigned char dummy2;
+			unsigned char dummy1;
+		} byte;
+	} src1;
+/*
+ *		[--------------longword----------------] SRC2_32
+ *		[                  ][------word--------] SRC2_16
+ *		[        ][        ][--byte--][--aux---]
+ *				      SRC2_8    AUX2_8
+ */
+	union {
+		unsigned long longword;
+		struct {
+			unsigned short word;
+			unsigned short dummy;
+		} word16;
+		struct {
+ 			unsigned char aux;
+			unsigned char byte;
+			unsigned char dummy2;
+			unsigned char dummy1;
+		} byte;
+	} src2;
+	unsigned char *reg1;	/* from REG field of mod_rm */
+	unsigned char *mem_ref;	/* from RM field of mod_rm */
+	int mode_reg;
+	unsigned long error_addr;
+	unsigned char bigseg[6];
 } Interp_ENV;
 #endif
 
@@ -1072,300 +1241,353 @@ typedef struct keyENV{
 	unsigned long   level;
 	struct HSW_86_CATCHBUF *buf;
 	jmp_buf		jump_buffer;
+	unsigned char *seg_regs[8];
+	Interp_VAR_flags_czsp flags_czsp;
+	union {
+		unsigned long longword;
+		struct {
+			unsigned short word;
+			unsigned short dummy;
+			unsigned int dummy1;
+		} word16;
+		struct {
+ 			unsigned char aux;
+			unsigned char byte;
+			unsigned char dummy2;
+			unsigned char dummy1;
+			unsigned int dummy3;
+		} byte;
+	} src1;
+	union {
+		unsigned long longword;
+		struct {
+			unsigned short word;
+			unsigned short dummy;
+			unsigned int dummy1;
+		} word16;
+		struct {
+ 			unsigned char aux;
+			unsigned char byte;
+			unsigned char dummy2;
+			unsigned char dummy1;
+			unsigned int dummy3;
+		} byte;
+	} src2;
+	unsigned char *reg1;	/* from REG field of mod_rm */
+	unsigned char *mem_ref;	/* from RM field of mod_rm */
+	int mode_reg;
 } Interp_ENV;
 #endif
 
-#if defined(sparc) || defined(mips) || defined(ppc) || defined(hppa)
-typedef struct {
-	unsigned char *seg_regs[7];
-	union {
-		unsigned long res;
-		struct {
-			unsigned short carry;
-			unsigned short res16;
-		} word16;
-		struct {
-			unsigned char byte_op;
-			unsigned char carryb;
-			unsigned char res8;
-			unsigned char parity16;
-		} byte;
-	} flags_czsp;
-	union {
-		unsigned long longword;
-		struct {
-			unsigned short dummy;
-			unsigned short word;
-		} word16;
-		struct {
-			unsigned char dummy1;
-			unsigned char dummy2;
-			unsigned char byte;
- 			unsigned char aux;
-		} byte;
-	} src1;
-	union {
-		unsigned long longword;
-		struct {
-			unsigned short dummy;
-			unsigned short word;
-		} word16;
-		struct {
-			unsigned char dummy1;
-			unsigned char dummy2;
-			unsigned char byte;
- 			unsigned char aux;
-		} byte;
-	} src2;
-	unsigned char *reg1;	/* from REG field of mod_rm */
-	unsigned char *mem_ref;	/* from RM field of mod_rm */
-	int mode_reg;
-} Interp_VAR;
-#endif
-
-#if defined(X386) 
-typedef struct {
-	unsigned char *seg_regs[7];
-	union {
-		unsigned long res;
-		struct {
-			unsigned short res16;
-			unsigned short carry;
-		} word16;
-		struct {
-			unsigned char parity16;
-			unsigned char res8;
-			unsigned char carryb;
-			unsigned char byte_op;
-		} byte;
-	} flags_czsp;
-	union {
-		unsigned long longword;
-		struct {
-			unsigned short word;
-			unsigned short dummy;
-		} word16;
-		struct {
- 			unsigned char aux;
-			unsigned char byte;
-			unsigned char dummy2;
-			unsigned char dummy1;
-		} byte;
-	} src1;
-	union {
-		unsigned long longword;
-		struct {
-			unsigned short word;
-			unsigned short dummy;
-		} word16;
-		struct {
- 			unsigned char aux;
-			unsigned char byte;
-			unsigned char dummy2;
-			unsigned char dummy1;
-		} byte;
-	} src2;
-	unsigned char *reg1;	/* from REG field of mod_rm */
-	unsigned char *mem_ref;	/* from RM field of mod_rm */
-	int mode_reg;
-} Interp_VAR;
-#endif
-
-#if defined(alpha)
-typedef struct {
-	unsigned char *seg_regs[7];
-	union {
-		unsigned long res;
-		struct {
-			unsigned short res16;
-			unsigned short carry;
-			unsigned int dummy1;
-		} word16;
-		struct {
-			unsigned char parity16;
-			unsigned char res8;
-			unsigned char carryb;
-			unsigned char byte_op;
-			unsigned int dummy1;
-		} byte;
-	} flags_czsp;
-	union {
-		unsigned long longword;
-		struct {
-			unsigned short word;
-			unsigned short dummy;
-			unsigned int dummy1;
-		} word16;
-		struct {
- 			unsigned char aux;
-			unsigned char byte;
-			unsigned char dummy2;
-			unsigned char dummy1;
-			unsigned int dummy3;
-		} byte;
-	} src1;
-	union {
-		unsigned long longword;
-		struct {
-			unsigned short word;
-			unsigned short dummy;
-			unsigned int dummy1;
-		} word16;
-		struct {
- 			unsigned char aux;
-			unsigned char byte;
-			unsigned char dummy2;
-			unsigned char dummy1;
-			unsigned int dummy3;
-		} byte;
-	} src2;
-	unsigned char *reg1;	/* from REG field of mod_rm */
-	unsigned char *mem_ref;	/* from RM field of mod_rm */
-	int mode_reg;
-} Interp_VAR;
-#endif
-
-typedef union {
-		int longw;
-		struct {
-			unsigned short snjr;
-			unsigned short jnr;
-		} word;
-} MULT;
-
-#ifndef DOSEMU_TYPESONLY
-
-#define CARRY interp_var->flags_czsp.word16.carry
-#define RES_32 interp_var->flags_czsp.res
-#define RES_16 interp_var->flags_czsp.word16.res16
-#define RES_8 interp_var->flags_czsp.byte.res8
-#define PAR_16 interp_var->flags_czsp.byte.parity16
-#define AUX1_8 interp_var->src1.byte.aux
-#define AUX2_8 interp_var->src2.byte.aux
-#define BYTE_FLAG interp_var->flags_czsp.byte.byte_op
-#define SRC1_32 interp_var->src1.longword
-#define SRC2_32 interp_var->src2.longword
-#define SRC1_16 interp_var->src1.word16.word
-#define SRC2_16 interp_var->src2.word16.word
-#define SRC1_8 interp_var->src1.byte.byte
-#define SRC2_8 interp_var->src2.byte.byte
-#define HREG1 (unsigned char *)(interp_var->reg1)
-#define XREG1 (unsigned short *)(interp_var->reg1)
-#define EREG1 (DWORD *)(interp_var->reg1)
-#define MEM_REF interp_var->mem_ref
-
-#define IS_MODE_REG interp_var->mode_reg
-
-#define ALLOW_OVERRIDE(default) ((interp_var->seg_regs[6]!=INVALID_OVR)?interp_var->seg_regs[6]:default)
-#define FETCH_XREG(addr) *(unsigned short *)addr
-#define FETCH_EREG(addr) *(DWORD *)addr
-#define PUT_XREG(addr,data) *(unsigned short *)addr=data
-#define PUT_EREG(addr,data) *(DWORD *)addr=data
-#ifdef DOSEMU
-#define IS_CF_SETD ((((src1 ^ src2) & (src1 ^ res)) | (src1 & src2)) >> 31)
-#else	/* buggy version */
-#define IS_CF_SETD ((((src1 ^ src2 ) & ~res) | (src1 & src2)) >> 31)
-#endif
-#define SETDFLAGS { DWORD res1=res;\
-RES_16 = ((res >> 24) << 8) | (res & 0xff); CARRY = IS_CF_SETD; BYTE_FLAG = 0; if(res1) RES_8 |= 0x1;\
-SRC1_16 = ((src1 >> 24) << 8) | (src1 & 0xff);\
-SRC2_16 = ((src2 >> 24) << 8) | (src2 & 0xff);}
-
-#if defined(DOSEMU) && defined(__i386__)
-#define	GETWORD(p)	(unsigned short)*((unsigned short *)(p))
-#define PUTWORD(p,w)	{*((unsigned short *)(p))=(unsigned short)(w);}
-#define PUTDWORD(p,dw)	{*((unsigned long *)(p))=(unsigned long)(dw);}
-#define GETDWORD(p)	(unsigned long)*((unsigned long *)(p))
+#ifdef USE_GLOBAL_EBP
+#define MODPARAMS	Interp_ENV *env
+#define MODARGS		env
 #else
-#include "Endian.h"
+#define MODPARAMS	Interp_ENV *env,BYTE *PC
+#define MODARGS		env,PC
 #endif
 
+#if !defined(DOSEMU) || defined(_IN_INTP32)
+
+/* CARRY is tricky, as it accesses also byte 3 (BYTE_OP); use with care */
+#define CARRY env->flags_czsp.word16.carry
+#define CARRYB env->flags_czsp.byte.carryb
+#define RES_32 env->flags_czsp.res
+#define RES_16 env->flags_czsp.word16.res16
+#define RES_8 env->flags_czsp.byte.res8
+#define AUXR_8 env->flags_czsp.byte.parity16
+#define PAR_16 env->flags_czsp.byte.parity16
+#define AUX1_8 env->src1.byte.aux
+#define AUX2_8 env->src2.byte.aux
+#define BYTE_FLAG env->flags_czsp.byte.byte_op
+#define SRC1_32 env->src1.longword
+#define SRC2_32 env->src2.longword
+#define SRC1_16 env->src1.word16.word
+#define SRC2_16 env->src2.word16.word
+#define SRC1_8 env->src1.byte.byte
+#define SRC2_8 env->src2.byte.byte
+/* these are pointers to the actual mem/reg src/dest location */
+#define HREG1 ((BYTE *)env->reg1)
+#define XREG1 ((WORD *)env->reg1)
+#define EREG1 ((DWORD *)env->reg1)
+#define MEM_REF env->mem_ref
+
+#define IS_MODE_REG env->mode_reg
+
+/*
+#define ALLOW_OVERRIDE(default) ((OVERRIDE!=INVALID_OVR)?OVERRIDE:default)
+*/
+
+/* ========================== FLAG setting ============================ */
+/* Use similar macros to save the op state for later flag evaluation -
+ * this greatly clarifies the source and avoids mistakes -- AV */
+/*
+ *		[-----------------res------------------]
+ *	     	[----------0000000c][--res(0)-----00---]
+ *		[01000101][        ][        ][        ]
+ *
+ * The parameter s is for the SUB flags setting (1) vs the ADD case (0).
+ * s should always be a compile-time constant to make gcc skip the code
+ * we don't need.
+ *
+ * SETBFLAGS requires definition of variables: src1,src2,res
+ */
+#ifdef __i386__
+/* no-carry version */
+#define BYTEDUP(sr)	({__asm__ __volatile__ ("\
+	movzbl	%b1,%0\n\
+	movb	%b1,%h0"\
+	: "=q"(sr) \
+	: "0"(sr) ); sr;})
+
+/* carry version */
+#define BYTEDUP_C(sr)	({__asm__ __volatile__ ("\
+	andl	$0x1ff,%0\n\
+	shll	$8,%0\n\
+	movb	%h1,%b0"\
+	: "=q"(sr) \
+	: "0"(sr) ); sr;})
+#else
+#define BYTEDUP(sr)	((((sr)&0xff)<<8)|((sr)&0xff))
+#define BYTEDUP_C(sr)	((((sr)&0x1ff)<<8)|((sr)&0xff))
+#endif
+
+/* New: align flag handling for byte ops to the default 16-bit word case.
+ * We do this by duplicating the bits 0..7 into 8..15. Carry goes always
+ * into bit 16; high byte is now cleared.
+ * Flags: SF: uses bit 15 as before
+ *	  ZF: if result byte is 0, so is the extended word
+ *	  OF: uses bit 15 as before. Low bytes of src1,src2 ignored
+ *	  PF: uses low byte, which is the same as the result
+ *	  CF: into bit 16 as before
+ *	  AF: uses low byte
+ */
+#define SETBFLAGS(s) { register DWORD _res1=res;\
+	RES_32=BYTEDUP_C(_res1);\
+	_res1=src1; SRC1_32=BYTEDUP(_res1);\
+	if (s) {_res1=-src2; if ((_res1&0x7f)==0) _res1^=0x80;}\
+	else _res1=src2; SRC2_32=BYTEDUP(_res1);}
+
+/*
+ *		[-----------------res------------------]
+ *	     	[----------0000000c][--res(1)---res(0)-]
+ *		[00000000][        ][        ][        ]
+ *
+ * The parameter s is for the SUB flags setting (1) vs the ADD case (0).
+ * s should always be a compile-time constant to make gcc skip the code
+ * we don't need.
+ *
+ * SETWFLAGS requires definition of variables: src1,src2,res
+ * High 16 bits of SRC1,SRC2 are don't cares
+ */
+#define SETWFLAGS(s) { register DWORD _res1;\
+	RES_32=res&0x1ffff; SRC1_32=src1;\
+	if (s) {_res1=-src2; if ((_res1&0x7fff)==0) _res1^=0x8000;}\
+	else _res1=src2; SRC2_32=_res1;}
+
+/* add/sub rule for carry using bit31:
+ *	src1 src2  res	b31(a) b31(s)
+ *	  0    0    0    0	0
+ *	  0    0    1    0	1
+ *	  0    1    0    1	1
+ *	  0    1    1    0	1
+ *	  1    0    0    1	0
+ *	  1    0    1    0	0
+ *	  1    1    0    1	0
+ *	  1    1    1    1	1
+ *
+ * This add/sub flag evaluation is tricky too. In 16-bit mode, the carry
+ * flag is always at its correct position (bit 16), while in 32-bit mode
+ * it has to be calculated from src1,src2,res. The original Willows code
+ * had the correct definition for the ADD case, but totally failed in the
+ * SUB/CMP case. It turned out that there isn't any simple expression
+ * covering both cases, BUT that the sub carry is the inverse of what we
+ * get for ADD if src2 is inverted (NOT negated!); see for yourself.
+ */ 
+#define IS_CF_SETD(s1,s2,r) ((unsigned)((((s1)^(s2)) & ~(r)) | ((s1)&(s2))) >> 31)
+
+/*
+ * for 32-bit ops we pack hi and lo bytes into a 16-bit value; this way
+ * we can use the same tests for sign (bit 15) and parity (lo byte). It
+ * is only necessary to add a case for the 32-bit zero condition; this is
+ * done by setting bit 8 if the original value was not zero, so that 16-bit
+ * zero test will give the correct answer -- AV
+ */
+#ifdef __i386__
+/* latest version... move only sign bit from b31 to b15; if sign=1, then
+ * of course ZF=0, else test and set bit 8 if 32-bit result is not zero.
+ * The trick works because rcr doesn't change the ZF coming from shl.
+ * Be careful not to change the LSByte - it is used by AF and PF */
+#define PACK32_16(sr,de)	__asm__ __volatile__ ("\
+	shll	%0\n\
+	rcrw	%w0\n\
+	jz	1f\n\
+	orb	$1,%h0\n\
+1:	movw	%w0,%w1"\
+	: "=q"(sr), "=g"(*((short *)&de))\
+	: "0"(sr) \
+	: "memory" )
+#else
+#define PACK32_16(sr,de)	de=((WORD)(((sr)>>16)&0xff00) |\
+				(WORD)((sr)&0xff) | (WORD)((sr)!=0? 0x100:0))
+#endif
+
+/*
+ *		[-----------------res------------------]
+ *	     	[----------sssssssc][--res(3)---res(0)-]
+ *		[00000000][        ][-------1][        ]
+ *
+ * There are two parameters for the flag setting macros; c controls if
+ * carry-setting flag has to be compiled in, s is for the SUB flags
+ * setting (1) vs the ADD case (0). c ad s should always be compile-time
+ * constants to make gcc skip the code we don't need.
+ *
+ * SETDFLAGS requires definition of variables: src1,src2,res
+ */
+#define SETDFLAGS(c,s) { register DWORD _res1=res;\
+	PACK32_16(_res1,RES_16);\
+	_res1=src1; PACK32_16(_res1,SRC1_16);\
+	if (s) {_res1=-src2; if ((_res1<<1)==0) _res1^=0x80000000; src2=~src2;}\
+	else _res1=src2; PACK32_16(_res1,SRC2_16);\
+	if (c) CARRY = IS_CF_SETD(src1,src2,res) ^ (s); else BYTE_FLAG=0;}
+
+
+/* ================== MEMORY and REGISTERS access ===================== */
+/*
+ * on x86, the CPU registers are stored in memory; there is no
+ * difference in accessing. We keep things separate for compatibility.
+ */
+
+#ifdef __i386__	/* ... and known little-endian machines */
+#define	GETBYTE(p)	(BYTE)*((BYTE *)(p))
+#define	GETWORD(p)	(WORD)*((WORD *)(p))
+#define GETDWORD(p)	(DWORD)*((DWORD *)(p))
+#define PUTBYTE(p,b)	{*((BYTE *)(p))=(BYTE)(b);}
+#define PUTWORD(p,w)	{*((WORD *)(p))=(WORD)(w);}
+#define PUTDWORD(p,dw)	{*((DWORD *)(p))=(DWORD)(dw);}
+#else	/* Twin: ugly but endian-independent */
+#define	GETWORD(p) (WORD)(((LPBYTE)(p))[0]+((WORD)(((LPBYTE)(p))[1])<<8))
+/* warning: problems with 'if..else' using brackets */
+#define PUTWORD(p,w) { *((LPBYTE)(p)) = (BYTE)((WORD)(w) % 0x100); \
+		       *((LPBYTE)(p) + 1) = (BYTE)((WORD)(w)>>8); }
+
+#define PUTDWORD(p,dw) { PUTWORD((LPBYTE)(p),LOWORD((DWORD)(dw))); \
+			PUTWORD((LPBYTE)(p)+2,HIWORD((DWORD)(dw))); }
+#define GETDWORD(p) ((DWORD)((DWORD)(GETWORD((LPBYTE)(p)+2)<<16) \
+			     + GETWORD((LPBYTE)(p))))
+#endif
+
+/* register definitions - can be different for some architectures */
+#define FETCH_HREG	GETBYTE
+#define FETCH_XREG	GETWORD
+#define FETCH_EREG	GETDWORD
+#define PUT_HREG	PUTBYTE
+#define PUT_XREG	PUTWORD
+#define PUT_EREG	PUTDWORD
+
+/* memory definitions */
+#define FETCH_BYTE	GETBYTE
 #define	FETCH_WORD	GETWORD
 #define	FETCH_QUAD	GETDWORD
+#define PUT_BYTE	PUTBYTE
 #define	PUT_WORD	PUTWORD
 #define	PUT_QUAD	PUTDWORD
 
+/* register/memory selection via macro */
 #ifdef DOSEMU
-#include "cpu-emu.h"
-#include "dosemu_debug.h"
-
-extern BOOL vm86f;
+#define mFETCH_BYTE	GETBYTE
+#define mFETCH_WORD	GETWORD
+#define mFETCH_QUAD	GETDWORD
+#define mPUT_BYTE	PUTBYTE
+#define mPUT_WORD	PUTWORD
+#define mPUT_QUAD	PUTDWORD
 #else
+#define mFETCH_BYTE(p)	(IS_MODE_REG? FETCH_HREG((p)):FETCH_BYTE((p)))
+#define mFETCH_WORD(p)	(IS_MODE_REG? FETCH_XREG((p)):FETCH_WORD((p)))
+#define mFETCH_QUAD(p)	(IS_MODE_REG? FETCH_EREG((p)):FETCH_QUAD((p)))
+#define mPUT_BYTE(p,b)	do{if(IS_MODE_REG) PUT_HREG((p),(b)); else PUT_BYTE((p),(b));}while(0)
+#define mPUT_WORD(p,w)	do{if(IS_MODE_REG) PUT_XREG((p),(w)); else PUT_WORD((p),(w));}while(0)
+#define mPUT_QUAD(p,dw)	do{if(IS_MODE_REG) PUT_EREG((p),(dw)); else PUT_QUAD((p),(dw));}while(0)
+#endif
+
+
+/* These are for multiply/divide. I ASSUME that every version of gcc on
+ * every 32-bit target has a 'long long' 64-bit type */
+typedef union {
+  QLONG sd;
+  struct { long sl,sh; } s;
+} s_i64_u;
+
+typedef union {
+  QWORD ud;
+  struct { long ul,uh; } u;
+} u_i64_u;
+
+#ifndef DOSEMU
 #define error(s...)	fprintf(stderr,##s)
-extern void INT_handler(int, ENV *);
+extern LPSTR	GetProcName(WORD wSel, WORD wOrd);
+#else
+#include "cpu-emu.h"
 #endif
-extern BOOL code32, data32;
-extern void FatalAppExit(UINT, LPCSTR);
 
 typedef void (*FUNCT_PTR)();
 
-#ifdef DEBUG
-extern void e_debug (Interp_ENV *env, unsigned char *P0, unsigned char *PC,
-  	Interp_VAR *interp_var);
+#ifndef NO_TRACE_MSGS
+extern void e_trace (Interp_ENV *env, unsigned char *, unsigned char *,
+  	int);
+extern void e_trace_fp (ENV87 *);
 #endif
+/* debug registers (DRs) emulation */
+extern int e_debug_check(unsigned char *);
 
 extern unsigned char *
-  hsw_interp_16_16 (Interp_ENV *env, unsigned char *P0, unsigned char *PC,
-  	Interp_VAR *interp_var, int *err);
+  hsw_interp_16_16 (Interp_ENV *env, unsigned char *,
+	unsigned char *, int *, int);
 extern unsigned char *
-  hsw_interp_16_32 (Interp_ENV *env, unsigned char *P0, unsigned char *PC,
-  	Interp_VAR *interp_var, int *err);
+  hsw_interp_16_32 (Interp_ENV *env, unsigned char *,
+	unsigned char *, int *);
 extern unsigned char *
-  hsw_interp_32_16 (Interp_ENV *env, unsigned char *P0, unsigned char *PC,
-  	Interp_VAR *interp_var, int *err);
+  hsw_interp_32_16 (Interp_ENV *env, unsigned char *,
+	unsigned char *, int *);
 extern unsigned char *
-  hsw_interp_32_32 (Interp_ENV *env, unsigned char *P0, unsigned char *PC,
-  	Interp_VAR *interp_var, int *err);
+  hsw_interp_32_32 (Interp_ENV *env, unsigned char *,
+	unsigned char *, int *, int);
 
 extern int
-  hsw_modrm_16_byte (Interp_ENV *env, unsigned char *PC, Interp_VAR *interp_var);
+  hsw_modrm_16_byte (MODPARAMS);
 extern int
-  hsw_modrm_16_word (Interp_ENV *env, unsigned char *PC, Interp_VAR *interp_var);
+  hsw_modrm_16_word (MODPARAMS);
 extern int
-  hsw_modrm_16_quad (Interp_ENV *env, unsigned char *PC, Interp_VAR *interp_var);
+  hsw_modrm_16_quad (MODPARAMS);
 extern int
-  hsw_modrm_32_byte (Interp_ENV *env, unsigned char *PC, Interp_VAR *interp_var);
+  hsw_modrm_32_byte (MODPARAMS);
 extern int
-  hsw_modrm_32_word (Interp_ENV *env, unsigned char *PC, Interp_VAR *interp_var);
+  hsw_modrm_32_word (MODPARAMS);
 extern int
-  hsw_modrm_32_quad (Interp_ENV *env, unsigned char *PC, Interp_VAR *interp_var);
+  hsw_modrm_32_quad (MODPARAMS);
 
-#ifdef DOSEMU
-#define e_AL env->rax.x.x.h.l
-#define e_AH env->rax.x.x.h.h
-#define e_AX env->rax.x.x.x
-#define e_EAX env->rax.e
-#define e_BL env->rbx.x.x.h.l
-#define e_BH env->rbx.x.x.h.h
-#define e_BX env->rbx.x.x.x
-#define e_EBX env->rbx.e
-#define e_CL env->rcx.x.x.h.l
-#define e_CH env->rcx.x.x.h.h
-#define e_CX env->rcx.x.x.x
-#define e_ECX env->rcx.e
-#define e_DL env->rdx.x.x.h.l
-#define e_DH env->rdx.x.x.h.h
-#define e_DX env->rdx.x.x.x
-#define e_EDX env->rdx.e
-#define e_SI env->rsi.si.si
-#define e_ESI env->rsi.esi
-#define e_DI env->rdi.di.di
-#define e_EDI env->rdi.edi
-#define e_BP env->rbp.bp.bp
-#define e_EBP env->rbp.ebp
-#define e_SP env->rsp.sp.sp
-#define e_ESP env->rsp.esp
-#define e_CS env->cs.cs
-#define e_DS env->ds.ds
-#define e_ES env->es.es
-#define e_SS env->ss.ss
-#define e_FS env->fs.fs
-#define e_GS env->gs.gs
-#define e_EIP env->trans_addr
-#define e_EFLAGS env->flags
+extern void invoke_data (Interp_ENV *);
 
-#endif	/* DOSEMU */
-#endif	/* DOSEMU_TYPESONLY */
+extern void
+trans_flags_to_interp(Interp_ENV *env, DWORD);
+extern DWORD
+trans_interp_flags(Interp_ENV *env);
+
+extern int
+init_cpuemu(unsigned long);
+
+#endif	/* _IN_INTP32 */
+
+extern BOOL code32, data32;
+extern void FatalAppExit(UINT, LPCSTR);
+extern QWORD EMUtime;
+
+#define MK_CS	0x00000
+#define MK_DS	0x10000
+#define MK_ES	0x20000
+#define MK_SS	0x30000
+#define MK_FS	0x40000
+#define MK_GS	0x50000
 
 /*
  * INT 00 C  - CPU-generated - DIVIDE ERROR
@@ -1406,4 +1628,66 @@ extern int
 #define EXCP11_ALGN	18
 #define EXCP12_MCHK	19
 
+#define EXCP_GOBACK	64
+#define EXCP_SIGNAL	65
+
+#if !defined(DOSEMU) || defined(_IN_INTP32)
+
+extern void hsw_fp87_00m(), hsw_fp87_01m(), hsw_fp87_02m(), hsw_fp87_03m();
+extern void hsw_fp87_04m(), hsw_fp87_05m(), hsw_fp87_06m(), hsw_fp87_07m();
+extern void hsw_fp87_10m(), hsw_fp87_11m(), hsw_fp87_12m(), hsw_fp87_13m();
+extern void hsw_fp87_14m(), hsw_fp87_15m(), hsw_fp87_16m(), hsw_fp87_17m();
+extern void hsw_fp87_20m(), hsw_fp87_21m(), hsw_fp87_22m(), hsw_fp87_23m();
+extern void hsw_fp87_24m(), hsw_fp87_25m(), hsw_fp87_26m(), hsw_fp87_27m();
+extern void hsw_fp87_30m(), hsw_fp87_31m(), hsw_fp87_32m(), hsw_fp87_33m();
+extern void hsw_fp87_34m(), hsw_fp87_35m(), hsw_fp87_36m(), hsw_fp87_37m();
+extern void hsw_fp87_40m(), hsw_fp87_41m(), hsw_fp87_42m(), hsw_fp87_43m();
+extern void hsw_fp87_44m(), hsw_fp87_45m(), hsw_fp87_46m(), hsw_fp87_47m();
+extern void hsw_fp87_50m(), hsw_fp87_51m(), hsw_fp87_52m(), hsw_fp87_53m();
+extern void hsw_fp87_54m(), hsw_fp87_55m(), hsw_fp87_56m(), hsw_fp87_57m();
+extern void hsw_fp87_60m(), hsw_fp87_61m(), hsw_fp87_62m(), hsw_fp87_63m();
+extern void hsw_fp87_64m(), hsw_fp87_65m(), hsw_fp87_66m(), hsw_fp87_67m();
+extern void hsw_fp87_70m(), hsw_fp87_71m(), hsw_fp87_72m(), hsw_fp87_73m();
+extern void hsw_fp87_74m(), hsw_fp87_75m(), hsw_fp87_76m(), hsw_fp87_77m();
+extern void hsw_fp87_00r(), hsw_fp87_01r(), hsw_fp87_02r(), hsw_fp87_03r();
+extern void hsw_fp87_04r(), hsw_fp87_05r(), hsw_fp87_06r(), hsw_fp87_07r();
+extern void hsw_fp87_10r(), hsw_fp87_11r(), hsw_fp87_12r(), hsw_fp87_13r();
+extern void hsw_fp87_14r(), hsw_fp87_15r(), hsw_fp87_16r(), hsw_fp87_17r();
+extern void hsw_fp87_20r(), hsw_fp87_21r(), hsw_fp87_22r(), hsw_fp87_23r();
+extern void hsw_fp87_24r(), hsw_fp87_25r(), hsw_fp87_26r(), hsw_fp87_27r();
+extern void hsw_fp87_30r(), hsw_fp87_31r(), hsw_fp87_32r(), hsw_fp87_33r();
+extern void hsw_fp87_34r(), hsw_fp87_35r(), hsw_fp87_36r(), hsw_fp87_37r();
+extern void hsw_fp87_40r(), hsw_fp87_41r(), hsw_fp87_42r(), hsw_fp87_43r();
+extern void hsw_fp87_44r(), hsw_fp87_45r(), hsw_fp87_46r(), hsw_fp87_47r();
+extern void hsw_fp87_50r(), hsw_fp87_51r(), hsw_fp87_52r(), hsw_fp87_53r();
+extern void hsw_fp87_54r(), hsw_fp87_55r(), hsw_fp87_56r(), hsw_fp87_57r();
+extern void hsw_fp87_60r(), hsw_fp87_61r(), hsw_fp87_62r(), hsw_fp87_63r();
+extern void hsw_fp87_64r(), hsw_fp87_65r(), hsw_fp87_66r(), hsw_fp87_67r();
+extern void hsw_fp87_70r(), hsw_fp87_71r(), hsw_fp87_72r(), hsw_fp87_73r();
+extern void hsw_fp87_74r(), hsw_fp87_75r(), hsw_fp87_76r(), hsw_fp87_77r();
+
+extern FUNCT_PTR hsw_fp87_mem0[8];
+extern FUNCT_PTR hsw_fp87_mem1[8];
+extern FUNCT_PTR hsw_fp87_mem2[8];
+extern FUNCT_PTR hsw_fp87_mem3[8];
+extern FUNCT_PTR hsw_fp87_mem4[8];
+extern FUNCT_PTR hsw_fp87_mem5[8];
+extern FUNCT_PTR hsw_fp87_mem6[8];
+extern FUNCT_PTR hsw_fp87_mem7[8];
+extern FUNCT_PTR hsw_fp87_reg0[8];
+extern FUNCT_PTR hsw_fp87_reg1[8];
+extern FUNCT_PTR hsw_fp87_reg2[8];
+extern FUNCT_PTR hsw_fp87_reg3[8];
+extern FUNCT_PTR hsw_fp87_reg4[8];
+extern FUNCT_PTR hsw_fp87_reg5[8];
+extern FUNCT_PTR hsw_fp87_reg6[8];
+extern FUNCT_PTR hsw_fp87_reg7[8];
+
+#endif
+
+extern BYTE parity[];
+extern char unknown_msg[], illegal_msg[], unsupp_msg[];
+extern ENV87 hsw_env87;
+
 #endif /* hsw_interp__h */
+
