@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -9,7 +9,7 @@
  *
  * Description: New port handling code for DOSEMU
  * 
- * Maintainers: Alberto Vignani (vignani@mbox.vol.it)
+ * Maintainers: Alberto Vignani (vignani@mail.tin.it)
  *
  * REMARK
  * This is the code that allows and disallows port access within DOSEMU.
@@ -46,6 +46,9 @@
 #endif
 #include <sys/stat.h>
 #include <sys/ioctl.h>
+#if GLIBC_VERSION_CODE >= 2000
+  #include <sys/io.h>
+#endif
 
 #include "emu.h"
 #include "port.h"
@@ -57,6 +60,11 @@
 #include "vc.h"
 #include "shared.h"
 #include "serial.h"
+#include "bitops.h"
+#ifdef X86_EMULATOR
+#include "cpu-emu.h"
+#include "bitops.h"
+#endif
 
 /*
  * maximum number of emulated devices allowed.  floppy, mda, etc...
@@ -117,7 +125,7 @@ void SET_HANDLE_COND(int p, int h)
   if (port_handle_table[(p)] < STD_HANDLES)
 	port_handle_table[(p)]=(h);
   else
-	T_printf("nPORT: %#x can't be mapped to handle %#x(%#x)\n",
+	i_printf("PORT: %#x can't be mapped to handle %#x(%#x)\n",
 	  p,h,port_handle_table[(p)]);
 }
 
@@ -143,100 +151,83 @@ static char *
 	return s + 3;
 }
 
-/* SIDOC_BEGIN_REMARK
- * VERB
- * PORT_DEBUG is to specify whether to record port writes to debug output.
- *   0 means disabled.
- *   1 means record all port accesses to 0x00 to 0xFF
- *   2 means record ANY port accesses!  (big fat debugfile!)
- *   3 means record all port accesses >= 0x100
- * /VERB
- * SIDOC_END_REMARK
- */
-#define PORT_DEBUG_READ	 2
-#define PORT_DEBUG_WRITE 2
-#define PORT_TRACE_D
+#define PORTLOG_MAXBITS		16
+#define PORTLOG_MASK		(~(-1 << PORTLOG_MAXBITS))
+#define SIZE_PORTLOGMAP		(1 << (PORTLOG_MAXBITS -3))
+static unsigned long *portlog_map = 0;
 
-#if PORT_DEBUG_READ > 0
-  static Bit8u log_port_read(ioport_t port, Bit8u r)
-  {
-    #if PORT_DEBUG_READ == 1
-      if (port < 0x100)
-    #elif PORT_DEBUG_READ == 3
-      if (port >= 0x100)
-    #endif
-      T_printf ("nPORTb: Rd 0x%04x -> 0x%02x %s\n", port, r, p2bin(r));
-      return r;
-  }
-  static Bit16u log_port_read_w(ioport_t port, Bit16u r)
-  {
-    #if PORT_DEBUG_READ == 1
-      if (port < 0x100)
-    #elif PORT_DEBUG_READ == 3
-      if (port >= 0x100)
-    #endif
-      T_printf ("PORTw: Rd 0x%04x -> 0x%04x\n", port, r);
-      return r;
-  }
-#ifdef PORT_TRACE_D
-  static Bit32u log_port_read_d(ioport_t port, Bit32u r)
-  {
-    #if PORT_DEBUG_READ == 1
-      if (port < 0x100)
-    #elif PORT_DEBUG_READ == 3
-      if (port >= 0x100)
-    #endif
-      T_printf ("PORTd: Rd 0x%04x -> 0x%08x\n", port, r);
-      return r;
-  }
-  #define LOG_PORT_READ_D(port, r) log_port_read_d(port, r)
-#endif
-  #define LOG_PORT_READ(port, r) log_port_read(port, r)
-  #define LOG_PORT_READ_W(port, r) log_port_read_w(port, r)
-#else
-  #define LOG_PORT_READ(port, r)	(r)
-  #define LOG_PORT_READ_W(port, r)	(r)
-  #define LOG_PORT_READ_D(port, r)	(r)
-#endif
+void init_port_traceing(void);
 
-#if PORT_DEBUG_WRITE > 0
-  static void log_port_write(ioport_t port, Bit8u w)
-  {
-    #if PORT_DEBUG_WRITE == 1
-      if (port < 0x100)
-    #elif PORT_DEBUG_WRITE == 3
-      if (port >= 0x100)
-    #endif
-      T_printf("PORTb: Wr 0x%04x <- %s 0x%02x\n", port, p2bin(w), w);
+void register_port_traceing(ioport_t firstport, ioport_t lastport)
+{
+  firstport &= PORTLOG_MASK;
+  lastport &= PORTLOG_MASK;
+  if (lastport < firstport) return;
+  init_port_traceing();
+  for (; firstport <= lastport; firstport++) {
+    set_bit(firstport, portlog_map);
   }
-  static void log_port_write_w(ioport_t port, Bit16u w)
-  {
-    #if PORT_DEBUG_WRITE == 1
-      if (port < 0x100)
-    #elif PORT_DEBUG_WRITE == 3
-      if (port >= 0x100)
-    #endif
-      T_printf("PORTw: Wr 0x%04x <- 0x%04x\n", port, w);
-  }
-#ifdef PORT_TRACE_D
-  static void log_port_write_d(ioport_t port, Bit32u w)
-  {
-    #if PORT_DEBUG_WRITE == 1
-      if (port < 0x100)
-    #elif PORT_DEBUG_WRITE == 3
-      if (port >= 0x100)
-    #endif
-      T_printf("PORTd: Wr 0x%04x <- 0x%08x\n", port, w);
-  }
-  #define LOG_PORT_WRITE_D(port, w) log_port_write_d(port, w)
-#endif
-  #define LOG_PORT_WRITE(port, w) log_port_write(port, w)
-  #define LOG_PORT_WRITE_W(port, w) log_port_write_w(port, w)
-#else
-  #define LOG_PORT_WRITE(port, w)
-  #define LOG_PORT_WRITE_W(port, w)
-  #define LOG_PORT_WRITE_D(port, w)
-#endif
+}
+
+void clear_port_traceing(void)
+{
+  if (!portlog_map) portlog_map = malloc(SIZE_PORTLOGMAP);
+  memset(portlog_map, 0, SIZE_PORTLOGMAP);
+}
+
+void init_port_traceing(void)
+{
+  if (portlog_map) return;
+  clear_port_traceing();
+  register_port_traceing(0x100, 0x2ff);
+}
+
+#define TT_printf(p,f,v,m) ({ \
+  if ( d.io_trace && test_bit(p & PORTLOG_MASK, portlog_map)) { \
+    log_printf(1, "%x %c %x\n", p, f, v & m); \
+  } \
+})
+
+static Bit8u log_port_read(ioport_t port, Bit8u r)
+{
+  TT_printf(port, '>', r, 0xff);
+  return r;
+}
+
+static Bit16u log_port_read_w(ioport_t port, Bit16u r)
+{
+  TT_printf(port, '}', r, 0xffff);
+  return r;
+}
+
+static Bit32u log_port_read_d(ioport_t port, Bit32u r)
+{
+  TT_printf(port, ']', r, 0xffffffff);
+  return r;
+}
+
+static void log_port_write(ioport_t port, Bit8u w)
+{
+  TT_printf(port, '<', w, 0xff);
+}
+
+static void log_port_write_w(ioport_t port, Bit16u w)
+{
+  TT_printf(port, '{', w, 0xffff);
+}
+
+static void log_port_write_d(ioport_t port, Bit32u w)
+{
+  TT_printf(port, '[', w, 0xffffffff);
+}
+
+#define LOG_PORT_READ(port, r) (d.io_trace ? log_port_read(port, r) : r)
+#define LOG_PORT_READ_W(port, r) (d.io_trace ? log_port_read_w(port, r) : r)
+#define LOG_PORT_READ_D(port, r) (d.io_trace ? log_port_read_d(port, r) : r)
+
+#define LOG_PORT_WRITE(port, w) do{ if (d.io_trace) log_port_write(port, w); }while(0)
+#define LOG_PORT_WRITE_W(port, w) do{ if (d.io_trace) log_port_write_w(port, w); }while(0)
+#define LOG_PORT_WRITE_D(port, w) do{ if (d.io_trace) log_port_write_d(port, w); }while(0)
 
 /* ---------------------------------------------------------------------- */
 /* SIDOC_BEGIN_REMARK
@@ -336,18 +327,12 @@ Bit32u port_ind(ioport_t port)
 	else {
 		res = std_port_ind(port);
 	}
-#ifdef PORT_TRACE_D
 	return LOG_PORT_READ_D(port, res);
-#else
-	return res;
-#endif
 }
 
 void port_outd(ioport_t port, Bit32u dword)
 {
-#ifdef PORT_TRACE_D
 	LOG_PORT_WRITE_D(port, dword);
-#endif
 	if (EMU_HANDLER(port).write_portd != NULL) {
 		EMU_HANDLER(port).write_portd(port, dword);
 	}
@@ -417,7 +402,7 @@ void std_port_outd(ioport_t port, Bit32u dword)
 
 static void pna_emsg(ioport_t port, char ch, char *s)
 {
-	T_printf("PORT%c: %x not available for %s\n", ch, port, s);
+	i_printf("PORT%c: %x not available for %s\n", ch, port, s);
 }
 
 static Bit8u port_not_avail_inb(ioport_t port)
@@ -472,9 +457,10 @@ int port_rep_inb(ioport_t port, Bit8u *base, int df, Bit32u count)
 {
 	register int incr = df? -1: 1;
 	Bit8u *dest = base;
+	int count_ = count;
 
 	if (count==0) return 0;
-	T_printf("Doing REP insb(%#x) %d bytes at %p, DF %d\n", port,
+	i_printf("Doing REP insb(%#x) %d bytes at %p, DF %d\n", port,
 		count, base, df);
 	if (EMU_HANDLER(port).read_portb == std_port_inb) {
 	    PORT_IOPLON(port,1);
@@ -490,6 +476,13 @@ int port_rep_inb(ioport_t port, Bit8u *base, int df, Bit32u count)
 	    dest += incr;
 	  }
 	}
+	if (d.io_trace) {
+		dest = base;
+		while (count_--) {
+			LOG_PORT_READ(port, *dest);
+			dest += incr;
+		}
+	}
 	return dest-base;
 }
 
@@ -497,9 +490,10 @@ int port_rep_outb(ioport_t port, Bit8u *base, int df, Bit32u count)
 {
 	register int incr = df? -1: 1;
 	Bit8u *dest = base;
+	int count_ = count;
 
 	if (count==0) return 0;
-	T_printf("Doing REP outsb(%#x) %d bytes at %p, DF %d\n", port,
+	i_printf("Doing REP outsb(%#x) %d bytes at %p, DF %d\n", port,
 		count, base, df);
 	if (EMU_HANDLER(port).write_portb == std_port_outb) {
 	    PORT_IOPLON(port,1);
@@ -515,6 +509,13 @@ int port_rep_outb(ioport_t port, Bit8u *base, int df, Bit32u count)
 	    dest += incr;
 	  }
 	}
+	if (d.io_trace) {
+		dest = base;
+		while (count_--) {
+			LOG_PORT_WRITE(port, *dest);
+			dest += incr;
+		}
+	}
 	return dest-base;
 }
 
@@ -522,9 +523,10 @@ int port_rep_inw(ioport_t port, Bit16u *base, int df, Bit32u count)
 {
 	register int incr = df? -1: 1;
 	Bit16u *dest = base;
+	int count_ = count;
 
 	if (count==0) return 0;
-	T_printf("Doing REP insw(%#x) %d words at %p, DF %d\n", port,
+	i_printf("Doing REP insw(%#x) %d words at %p, DF %d\n", port,
 		count, base, df);
 	if (EMU_HANDLER(port).read_portw == std_port_inw) {
 	    PORT_IOPLON(port,2);
@@ -548,6 +550,13 @@ int port_rep_inw(ioport_t port, Bit16u *base, int df, Bit32u count)
 	    dest += incr;
 	  }
 	}
+	if (d.io_trace) {
+		dest = base;
+		while (count_--) {
+			LOG_PORT_READ_W(port, *dest);
+			dest += incr;
+		}
+	}
 	return (Bit8u *)dest-(Bit8u *)base;
 }
 
@@ -555,9 +564,10 @@ int port_rep_outw(ioport_t port, Bit16u *base, int df, Bit32u count)
 {
 	register int incr = df? -1: 1;
 	Bit16u *dest = base;
+	int count_ = count;
 
 	if (count==0) return 0;
-	T_printf("Doing REP outsw(%#x) %d words at %p, DF %d\n", port,
+	i_printf("Doing REP outsw(%#x) %d words at %p, DF %d\n", port,
 		count, base, df);
 	if (EMU_HANDLER(port).write_portw == std_port_outw) {
 	    PORT_IOPLON(port,2);
@@ -581,6 +591,13 @@ int port_rep_outw(ioport_t port, Bit16u *base, int df, Bit32u count)
 	    dest += incr;
 	  }
 	}
+	if (d.io_trace) {
+		dest = base;
+		while (count_--) {
+			LOG_PORT_WRITE_W(port, *dest);
+			dest += incr;
+		}
+	}
 	return (Bit8u *)dest-(Bit8u *)base;
 }
 
@@ -592,6 +609,7 @@ int port_rep_ind(ioport_t port, Bit32u *base, int df, Bit32u count)
 	if (count==0) return 0;
 	while (count--) {
 	  *dest = port_ind(port);
+	  LOG_PORT_READ_D(port, *dest);
 	  dest += incr;
 	}
 	return (Bit8u *)dest-(Bit8u *)base;
@@ -605,6 +623,7 @@ int port_rep_outd(ioport_t port, Bit32u *base, int df, Bit32u count)
 	if (count==0) return 0;
 	while (count--) {
 	  port_outd(port, *dest);
+	  LOG_PORT_WRITE_D(port, *dest);
 	  dest += incr;
 	}
 	return (Bit8u *)dest-(Bit8u *)base;
@@ -618,11 +637,22 @@ int port_rep_outd(ioport_t port, Bit32u *base, int df, Bit32u count)
  * I don't know what to do of this stuff... it was added incrementally to
  * port.c and has mainly to do with video code. This is not the right
  * place for it...
- * Anyway, this implements some HGC stuff for X
+ * Anyway, this implements some HGC stuff for X and the emuretrace
+ * port access for 0x3c0/0x3da
  *
  * SIDOC_END_FUNCTION
  */
 
+static int r3da_pending = 0;
+
+void do_r3da_pending (void)
+{
+  if (r3da_pending) {
+    (void)std_port_inb(r3da_pending);
+    r3da_pending = 0;
+  }
+}
+
 static Bit8u special_port_inb(ioport_t port)
 {
 	Bit8u res = 0xff;
@@ -639,12 +669,14 @@ static Bit8u special_port_inb(ioport_t port)
 		return ((res & 0xfd) | (hgc_Konv & 0x02));
 	    }
 	}
-#if X_GRAPHICS
 	if ((port==0x3ba)||(port==0x3da)) {
-		res = Attr_get_input_status_1();
+		res = Misc_get_input_status_1();
+		if (!config.usesX && !r3da_pending && (config.emuretrace>1)) {
+			set_ioperm(0x3c0,1,0);
+			r3da_pending = port;
+		}
 	}
 	else
-#endif
 	if (port==0x3db)	/* light pen strobe reset */
 		res = 0;
 	return res;
@@ -674,6 +706,32 @@ static void special_port_outb(ioport_t port, Bit8u byte)
 		static int last_index = -1;
 		static int flip_flop = 1;
 
+		/* SIDOC_BEGIN_REMARK
+		 *
+		 * This is the core of the new emuretrace algorithm:
+		 * If a read of port 0x3da is performed we just set it
+		 *  as pending and set ioperm OFF for port 0x3c0
+		 * When a write to port 0x3c0 is then trapped, we perform
+		 *  any pending read to 0x3da and reset the ioperm for
+		 *  0x3c0 in the default ON state.
+		 * This way we avoid extra port accesses when the program
+		 * is only looking for the sync bits, and we don't miss
+		 * the case where the read to 0x3da is used to reset the
+		 * index/data flipflop for port 0x3c0. Futher accesses to
+		 * port 0x3c0 are handled at full speed.
+		 *
+		 * SIDOC_END_REMARK
+		 */
+		if (config.vga && !config.usesX && (config.emuretrace>1)) {
+		    if (r3da_pending) {
+			(void)std_port_inb(r3da_pending);
+			r3da_pending = 0;
+			set_ioperm(0x3c0,1,1);
+			port_real_outb(0x3c0, byte);
+			return;
+		    }
+		    goto defout;
+		}
 		flip_flop = !flip_flop;
 		if (flip_flop) {
 		/* JES This was last_index = 0x10..... WRONG? */
@@ -695,7 +753,7 @@ static void special_port_outb(ioport_t port, Bit8u byte)
 		static unsigned int hi = 0, lo = 0;
 		int pos;
 
-		v_printf("nPORT: 6845 outb [0x%04x]\n", port);
+		v_printf("PORT: 6845 outb [0x%04x]\n", port);
 		if ((port == READ_WORD(BIOS_VIDEO_PORT) + 1) && (last_port == READ_WORD(BIOS_VIDEO_PORT))) {
 			/* We only take care of cursor positioning for now. */
 			/* This code should work most of the time, but can
@@ -748,6 +806,9 @@ int port_init(void)
 	  port_handler[i].irq = EMU_NO_IRQ;
 	  port_handler[i].fd = -1;
 	}
+#ifdef X86_EMULATOR
+	memset (io_bitmap, 0, sizeof(io_bitmap));
+#endif
 
   /* handle 0 maps to the unmapped IO device handler.  Basically any
      ports which don't map to any other device get mapped to this
@@ -900,8 +961,10 @@ int extra_port_init(void)
 	SET_HANDLE_COND(0x3db,HANDLE_SPECIAL);		/* R */
 
 	i = READ_WORD(BIOS_VIDEO_PORT);
-	SET_HANDLE_COND(i,HANDLE_SPECIAL);		/* W */
-	SET_HANDLE_COND(i+1,HANDLE_SPECIAL);		/* W */
+	if (i) {	/* !config.vga */
+	    SET_HANDLE_COND(i,HANDLE_SPECIAL);		/* W */
+	    SET_HANDLE_COND(i+1,HANDLE_SPECIAL);	/* W */
+	}
 
  	return 0;
 }
@@ -939,7 +1002,7 @@ int port_register_handler(emu_iodev_t device, int flags)
     struct stat devstat;
 
     if (device.irq != EMU_NO_IRQ && device.irq >= EMU_MAX_IRQS) {
-	dbug_printf("nPORT: IO device %s registered with IRQ=%d above %u\n",
+	dbug_printf("PORT: IO device %s registered with IRQ=%d above %u\n",
 	      device.handler_name, device.irq, EMU_MAX_IRQS - 1);
 	return 1;
     }
@@ -954,12 +1017,12 @@ int port_register_handler(emu_iodev_t device, int flags)
     if (handle >= port_handles) {
 	/* no existing handle found, create new one */
 	if (port_handles >= EMU_MAX_IO_DEVICES) {
-		error("nPORT: too many IO devices, increase EMU_MAX_IO_DEVICES");
+		error("PORT: too many IO devices, increase EMU_MAX_IO_DEVICES");
 		leavedos(77);
 	}
 
 	if (device.irq != EMU_NO_IRQ && irq_handler_name[device.irq]) {
-		error("nPORT: IRQ %d conflict.  IO devices %s & %s\n",
+		error("PORT: IRQ %d conflict.  IO devices %s & %s\n",
 		      device.irq, irq_handler_name[device.irq], device.handler_name);
 		if (device.fd) close(device.fd);
 		return 2;
@@ -991,7 +1054,7 @@ int port_register_handler(emu_iodev_t device, int flags)
   /* change table to reflect new handler id for that address */
     for (i = device.start_addr; i <= device.end_addr; i++) {
 	if (port_handle_table[i] != 0) {
-		error("nPORT: conflicting devices: %s & %s\n",
+		error("PORT: conflicting devices: %s & %s\n",
 		      port_handler[handle].handler_name, EMU_HANDLER(i).handler_name);
 		if (device.fd) close(device.fd);
 		return 4;
@@ -999,18 +1062,18 @@ int port_register_handler(emu_iodev_t device, int flags)
 	port_handle_table[i] = handle;
     }
 
-    T_printf("nPORT: registered \"%s\" handle 0x%02x [0x%04x-0x%04x] fd=%d\n",
+    i_printf("PORT: registered \"%s\" handle 0x%02x [0x%04x-0x%04x] fd=%d\n",
 	port_handler[handle].handler_name, handle, device.start_addr,
 	device.end_addr, (device.fd>=0? devstat.st_dev:device.fd));
 
     if (flags & PORT_FAST) {
 	if (device.start_addr < 0x400) {
-	  T_printf("nPORT: giving fast access to ports [0x%04x-0x%04x]\n",
+	  i_printf("PORT: giving fast access to ports [0x%04x-0x%04x]\n",
 		device.start_addr, device.end_addr);
 	  set_ioperm (device.start_addr, device.end_addr-device.start_addr+1, 1);
 	}
 	else
-	  T_printf("nPORT: using perm/iopl for ports [0x%04x-0x%04x]\n",
+	  i_printf("PORT: using perm/iopl for ports [0x%04x-0x%04x]\n",
 		device.start_addr, device.end_addr);
     }
     return 0;
@@ -1034,14 +1097,14 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 	char *devrname;
 	int fd, usemasks = 0;
 
-	T_printf("nPORT: allow_io for port 0x%04x:%d perm=%x or=%x and=%x\n",
+	i_printf("PORT: allow_io for port 0x%04x:%d perm=%x or=%x and=%x\n",
 		 start, size, permission, ormask, andmask);
 
 	if ((ormask != 0) || (andmask != 0xff)) {
 		if ((start+size) > 0x400)
-			T_printf("nPORT: andmask & ormask not supported for ports>=0x400\n");
+			i_printf("PORT: andmask & ormask not supported for ports>=0x400\n");
 		else if (size > 1)
-			T_printf("nPORT: andmask & ormask not supported for multiple ports\n");
+			i_printf("PORT: andmask & ormask not supported for multiple ports\n");
 		else
 			usemasks = 1;
 	}
@@ -1058,7 +1121,7 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 	 * SIDOC_END_REMARK
 	 */
 	if ((fp = fopen("/proc/ioports", "r")) == NULL) {
-		T_printf("nPORT: can't open /proc/ioports\n");
+		i_printf("PORT: can't open /proc/ioports\n");
 		return FALSE;
 	}
 	mapped = 0;
@@ -1066,7 +1129,7 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 		sscanf(line, "%x-%x : %s", &beg, &end, portname);
 		if ((start <= end) && ((start+size) > beg)) {
 			/* ports are besetzt, try to open the according device */
-			T_printf("nPORT: in range 0x%04x-0x%04x already registered as %s\n",
+			i_printf("PORT: in range 0x%04x-0x%04x already registered as %s\n",
 				 beg, end, portname);
 			if (!strncasecmp(portname,"dosemu",6)) return FALSE;
 			mapped = 1;
@@ -1076,7 +1139,7 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 	fclose (fp);
 
 	if (mapped && ((device==NULL) || (*device==0))) {
-		T_printf ("nPORT: no device specified for %s\n", portname);
+		i_printf ("PORT: no device specified for %s\n", portname);
 		return FALSE;
 	}
 
@@ -1116,17 +1179,17 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 		if (io_device.fd == -1) {
 			switch (errno) {
 			case EBUSY:
-				T_printf("nPORT: Device %s busy\n", device);
+				i_printf("PORT: Device %s busy\n", device);
 				return FALSE;
 			case EACCES:
-				T_printf("nPORT: Device %s, access not allowed\n", device);
+				i_printf("PORT: Device %s, access not allowed\n", device);
 				return FALSE;
 			case ENOENT:
 			case ENXIO:
-				T_printf("nPORT: No such Device '%s'\n", device);
+				i_printf("PORT: No such Device '%s'\n", device);
 				return FALSE;
 			default:
-				T_printf("nPORT: Device %s error %d\n", device, errno);
+				i_printf("PORT: Device %s error %d\n", device, errno);
 				return FALSE;
 			}
 		}
@@ -1134,11 +1197,11 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 		fd = open(lock_file, O_RDONLY);
 		if (fd >= 0) {
 			close(fd);
-			T_printf("nPORT: Device %s is locked\n", device);
+			i_printf("PORT: Device %s is locked\n", device);
 			return FALSE;
 		}
 
-		T_printf("nPORT: Device %s opened successfully = %d\n", device,
+		i_printf("PORT: Device %s opened successfully = %d\n", device,
 			io_device.fd);
 
 	}
@@ -1188,8 +1251,15 @@ set_ioperm(int start, int size, int flag)
 	tmp = DOS_SYSCALL(ioperm(start, size, flag));
 	leave_priv_setting();
 
-	T_printf ("nPORT: set_ioperm [%4x:%2d:%d] returns %d\n",start,size,flag,tmp);
-
+	i_printf ("nPORT: set_ioperm [%4x:%2d:%d] returns %d\n",start,size,flag,tmp);
+#ifdef X86_EMULATOR
+	if (config.cpuemu && (tmp==0)) {
+	    int i;
+	    for (i=start; i<(start+size); i++)
+		(flag? set_bit(i,io_bitmap) : clear_bit(i,io_bitmap));
+	    i_printf("ePORT: set_ioperm [%4x:%2d:%d]\n",start,size,flag);
+	}
+#endif
 	return tmp;
 }
 
