@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -41,7 +41,7 @@
  */
 
 #ifdef __NetBSD__
-#define __ELF__				/* simluated with a.out mmap-ing stuff.
+#define __ELF__				/* simulated with a.out mmap-ing stuff.
 					   use _main entrypoint. */
 #define EDEADLOCK EDEADLK
 #endif
@@ -89,7 +89,7 @@ __asm__("___START___: jmp _emulate\n");
 #include "netbsd_vm86.h"
 #endif
 #ifdef __linux__
-#if __GLIBC__ > 1
+#if GLIBC_VERSION_CODE >= 2000
 #include <sys/vt.h>
 #include <sys/kd.h>
 #else
@@ -131,9 +131,6 @@ __asm__("___START___: jmp _emulate\n");
 #include "priv.h"   /* for priv_init */
 #include "port.h"   /* for port_init */
 #include "pci.h"
-#ifdef __NetBSD__
-#include <setjmp.h>
-#endif
 #include "speaker.h"
 #include "utilities.h"
 #include "dos2linux.h"
@@ -143,6 +140,9 @@ __asm__("___START___: jmp _emulate\n");
 #ifdef USE_SBEMU
 #include "sound.h"
 #endif
+#ifdef X86_EMULATOR
+#include "cpu-emu.h"
+#endif
 
 extern void     stdio_init(void);
 extern void     time_setting_init(void);
@@ -163,7 +163,7 @@ extern void io_select_init(void);
 extern void treads_init(void);
 #endif
 
-
+jmp_buf NotJEnv;
 
 void 
 boot(void)
@@ -382,19 +382,30 @@ changesegs()
  * 
  */
 #ifdef __ELF__
-void 
+int 
 main(int argc, char **argv)
 #else
-void 
+int
 emulate(int argc, char **argv)
 #endif
 {
     extern void parse_dosemu_users(void);
+    int e;
 #ifdef __NetBSD__
     changesegs();
 #endif
 
     srand(time(NULL));
+    memset(&config, 0, sizeof(config));
+
+    if ((e=setjmp(NotJEnv))) {
+        flush_log();
+    	fprintf(stderr,"EMERGENCY JUMP %x!!!\n",e);
+    	/* there's no other way to stop a signal 11 from hanging dosemu
+    	 * but politely ask the kernel to terminate ourselves */
+	kill(0, SIGKILL);
+	_exit(1);		/* just in case */
+    }
 
 #ifdef USE_THREADS
     treads_init();		/* init the threads system,
@@ -423,6 +434,10 @@ emulate(int argc, char **argv)
     port_init();		/* setup port structures, before config! */
     version_init();		/* Check the OS version */
     config_init(argc, argv);	/* parse the commands & config file(s) */
+#ifdef DONT_DEBUG_BOOT		/* cpuemu only */
+    memcpy(&d_save,&d,sizeof(struct debug_flags));
+    if (d.emu) memset(&d,0,sizeof(struct debug_flags));
+#endif
     get_time_init();
     stdio_init();		/* initialize stdio & open debug file */
     print_version();            /* log version information */
@@ -436,6 +451,10 @@ emulate(int argc, char **argv)
     hardware_setup();		/* setup any hardware */
     extra_port_init();		/* setup ports dependent on config */
     memory_init();		/* initialize the memory contents */
+
+    /* here we include the hooks to possible plug-ins */
+    #include "plugin_init.h"
+
     boot();			/* read the boot sector & get moving */
 
     if (not_use_sigio)
@@ -453,6 +472,8 @@ emulate(int argc, char **argv)
 
     while (!fatalerr) {
 	++pic_vm86_count;
+	if (d.general>6)
+	  g_printf("------ EMU: main loop -- %06ld -----------\n",pic_vm86_count);
 	run_vm86();
 #if 0
 	timer_int_engine();
@@ -481,6 +502,7 @@ emulate(int argc, char **argv)
     sync();
     fprintf(stderr, "Not a good day to die!!!!!\n");
     leavedos(99);
+    return 0;  /* just to make gcc happy */
 }
 
 #if 0    /* disable C-A-D until someone will fix it (if really needed) */
@@ -520,7 +542,7 @@ ign_sigs(int sig)
     static int      timerints = 0;
     static int      otherints = 0;
 
-    g_printf("ERROR: signal %d received in leavedos()\n", sig);
+    error("signal %d received in leavedos()\n", sig);
     show_regs(__FILE__, __LINE__);
     flush_log();
     if (sig == SIG_TIME)
@@ -530,9 +552,9 @@ ign_sigs(int sig)
 
 #define LEAVEDOS_TIMEOUT (3 * config.freq)
 #define LEAVEDOS_SIGOUT  5
-    if ((timerints >= LEAVEDOS_TIMEOUT) || (otherints >= LEAVEDOS_SIGOUT)) {
+    if ((sig==11) || (timerints >= LEAVEDOS_TIMEOUT) || (otherints >= LEAVEDOS_SIGOUT)) {
 	error("timed/signalled out in leavedos()\n");
-	exit(1);
+	longjmp(NotJEnv, 0x11);
     }
 }
 
@@ -546,13 +568,16 @@ leavedos(int sig)
     struct sigaction sa;
     struct itimerval itv;
     extern int errno;
-
+    extern void do_r3da_pending (void);	/* emuretrace stuff */
+#if defined(X86_EMULATOR) && defined(EMU_STAT)
+    extern void print_emu_stat(void);
+#endif
    
     if (leavedos_recurse_check)
       {
        error("leavedos called recursively, forgetting the graceful exit!\n");
        flush_log();
-       _exit(sig);
+       longjmp(NotJEnv, sig);
       }
     leavedos_recurse_check = 1;
     warn("leavedos(%d) called - shutting down\n", sig);
@@ -582,6 +607,9 @@ leavedos(int sig)
     SETSIG(SIGTRAP, ign_sigs);
     warn("leavedos(%d) called - shutting down\n", sig);
 
+    /* here we include the hooks to possible plug-ins */
+    #include "plugin_close.h"
+
     if (config.speaker == SPKR_EMULATED) {
       g_printf("SPEAKER: sound off\n");
       speaker_off();		/* turn off any sound */
@@ -591,50 +619,51 @@ leavedos(int sig)
        /* Since the speaker is native hardware use port manipulation,
 	* we don't know what is actually implementing the kernel's
 	* ioctls.
-	* My port logic is acutally stolen from kd_nosound in the kernel.
+	* My port logic is actually stolen from kd_nosound in the kernel.
 	* 		--EB 21 September 1997
 	*/
        port_safe_outb(0x61, port_safe_inb(0x61)&0xFC); /* turn off any sound */
     }
 
-    g_printf("calling close_all_printers\n");
-    close_all_printers();
-
-    g_printf("releasing ports and blocked devices\n");
-    release_ports();
-
-    g_printf("calling serial_close\n");
-    serial_close();
-    g_printf("calling mouse_close\n");
-    mouse_close();
-
-#ifdef IPX
-    {
-      extern void ipx_close(void);
-      ipx_close();
-    }
-#endif
-
 #ifdef SIG
     g_printf("calling SIG_close\n");
 #endif
     SIG_close();
 
+    /* try to regain control of keyboard and video first */
+    g_printf("calling keyboard_close\n");
+    keyb_server_close();
+    keyb_client_close();
+
+#if defined(X86_EMULATOR) && defined(EMU_STAT)
+    /* if we are here with config.cpuemu>1 something went wrong... */
+    if (config.cpuemu>1) { d.emu=1; print_emu_stat(); d.emu=config.cpuemu=0; }
+#else
     show_ints(0, 0x33);
+#endif
     g_printf("calling disk_close_all\n");
     disk_close_all();
     g_printf("calling video_close\n");
     video_close();
    
-    g_printf("calling keyboard_close\n");
-    keyb_server_close();
-    keyb_client_close();
+    if (config.emuretrace) {
+      do_r3da_pending ();
+      set_ioperm (0x3da, 1, 1);
+      set_ioperm (0x3c0, 1, 1);
+      config.emuretrace = 0;
+    }
 
+    g_printf("releasing ports and blocked devices\n");
+    release_ports();
 
     g_printf("calling shared memory exit\n");
     shared_memory_exit();
     g_printf("calling HMA exit\n");
     hma_exit();
+    {
+	extern void close_uhook(void);
+	close_uhook();
+    }
 #ifdef USE_MHPDBG
     g_printf("closing debugger pipes\n");
     mhp_close();
@@ -643,12 +672,31 @@ leavedos(int sig)
 	restore_vt(config.detach);
 	disallocate_vt();
     }
+
+#ifdef IPX
+    {
+      extern void ipx_close(void);
+      ipx_close();
+    }
+#endif
+
+    g_printf("calling close_all_printers\n");
+    close_all_printers();
+
+    g_printf("calling serial_close\n");
+    serial_close();
+    g_printf("calling mouse_close\n");
+    mouse_close();
+
     flush_log();
 
     /* remove per process tmpdir and its contents */
     {
        char *command = strcatdup("rm -rf >/dev/null 2>&1 ", TMPDIR_PROCESS);
-       if (command) run_unix_command(command); /* beware of system() !! */
+       if (command) {
+          priv_drop(); /* drop any priviledges before running system() !! */
+          system(command);
+       }
     }
 
 #ifdef USE_THREADS
