@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -13,12 +13,9 @@
 int X_change_config(unsigned, void *);		/* modify X config data from DOS */
 
 void           get_vga_colors (void);
-int            X_init         (void);
-void           X_close        (void);
-void           X_redraw_screen (void);
-int            X_update_screen (void);
 void           X_handler      (void);
 void X_draw_cursor(int x,int y);
 void X_restore_cell(int x,int y);
 void X_move_cursor(int from_x,int from_y,int to_x,int to_y);
 void X_setcursorshape(unsigned short shape);
+void X_set_textsize(int, int);
