@@ -1,6 +1,6 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -47,6 +47,9 @@
 #define MODE_VGA_2	(1 << 10)
 #define MODE_VGA_4	(1 << 11)
 #define MODE_VGA_X	(1 << 12)
+#define MODE_CGA_1	(1 << 13)
+#define MODE_CGA_2	(1 << 14)
+#define MODE_HERC	(1 << 15)
 #define MODE_UNSUP	(1 << 31)
 
 #define MODE_TRUE_COL	(MODE_TRUE_8 | MODE_TRUE_15 | MODE_TRUE_16 | MODE_TRUE_24 | MODE_TRUE_32)
