@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -12,18 +12,23 @@
 #include "config.h"
 #include "extern.h"
 
-#if __GNUC__ >= 2
+#if GCC_VERSION_CODE >= 2005
 # define FORMAT(T,A,B)  __attribute__((format(T,A,B)))
 #else
 # define FORMAT(T,A,B)
 #endif
 
-#if __GNUC__ >= 2
+#if GCC_VERSION_CODE >= 2005
 # define NORETURN	__attribute__((noreturn))
 #else
 # define NORETURN
 #endif
 
+#ifdef X86_EMULATOR
+#if 1
+#define DONT_DEBUG_BOOT
+#endif
+#endif
 
 /* "dRWDCvXkiTsm#pgcwhIExMnPrS" */
 struct debug_flags {
@@ -53,13 +58,17 @@ struct debug_flags {
    network,		/* IPX network       "n" */
    pd,			/* Packet driver     "P" */
    request,		/* PIC               "r" */
-   sound 		/* SOUND	     "S" */
+   sound, 		/* SOUND	     "S" */
+   aspi			/* ASPI interface    "A" */
 #ifdef X86_EMULATOR
    ,emu			/* CPU emulation     "e" */
 #endif
    ;
 };
 
+#ifdef DONT_DEBUG_BOOT
+extern struct debug_flags d_save;
+#endif
 extern struct debug_flags d;
 
 int log_printf(int, const char *,...) FORMAT(printf, 2, 3);
@@ -119,6 +128,7 @@ void verror(const char *fmt, va_list args);
 #define pd_printf(f,a...)       ifprintf(d.pd,f,##a)	   /* pktdrvr */
 #define r_printf(f,a...)        ifprintf(d.request,f,##a)
 #define S_printf(f,a...)     	ifprintf(d.sound,f,##a)
+#define A_printf(f,a...)     	ifprintf(d.aspi,f,##a)
 #ifdef X86_EMULATOR
 #define e_printf(f,a...)     	ifprintf(d.emu,f,##a)
 #endif
@@ -154,6 +164,7 @@ void verror(const char *fmt, va_list args);
 #define pd_printf(f,a...)
 #define r_printf(f,a...)
 #define S_printf(f,a...)
+#define A_printf(f,a...)
 #define e_printf(f,a...)
 
 #define ALL_DEBUG_ON
