@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -14,10 +14,11 @@
 
 #include <features.h>
 #include <sys/types.h>
-#include <signal.h> 
-#if __GLIBC__ == 2 && __GLIBC_MINOR__ == 0
+#include <setjmp.h>
+#if GLIBC_VERSION_CODE == 2000
 #include <sigcontext.h>
 #endif
+#include <signal.h> 
 #include "config.h"
 #include "types.h"
 #include "extern.h"
@@ -71,6 +72,11 @@ EXTERN int screen, screen_mode;
 /* number of highest vid page - 1 */
 EXTERN int max_page INIT(7);
 
+#if 0
+/*
+ * 1) this stuff is unused
+ * 2) it should be FORBIDDEN to use global names less than 4 chars long!
+ */
 EXTERN char *cl,		/* clear screen */
 *le,				/* cursor left */
 *cm,				/* goto */
@@ -87,6 +93,7 @@ EXTERN char *cl,		/* clear screen */
 *ke,				/* ens keys */
 *vi,				/* cursor invisible */
 *ve;				/* cursor normal */
+#endif
 
 /* the fd for the keyboard */ 
 EXTERN int console_fd INIT(-1);
@@ -100,6 +107,10 @@ EXTERN int keypipe;
 EXTERN int mousepipe;
 
 EXTERN int in_vm86 INIT(0);
+#ifdef X86_EMULATOR
+EXTERN int in_vm86_emu INIT(0);
+EXTERN int in_dpmi_emu INIT(0);
+#endif
 
 EXTERN int li, co;	/* lines, columns */
 EXTERN int scanseq;
@@ -109,10 +120,13 @@ EXTERN int cursor_col;
 #if 0
 void dos_ctrl_alt_del(void);	/* disabled */
 #endif
+extern jmp_buf NotJEnv;
 
 EXTERN void run_vm86(void);
 EXTERN void     vm86_GP_fault();
 
+EXTERN void do_call_back(Bit32u codefarptr);
+
 #define NOWAIT  0
 #define WAIT    1
 #define TEST    2
@@ -140,6 +154,9 @@ EXTERN struct debug_flags d INIT({0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 #else
 EXTERN struct debug_flags d INIT({0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0});
 #endif
+#ifdef DONT_DEBUG_BOOT
+EXTERN struct debug_flags d_save;
+#endif
  
 EXTERN u_char in_sighandler, in_ioctl;
 /* one-entry queue ;-( for ioctl's */
@@ -285,6 +302,7 @@ typedef struct vesamode_type_struct {
        boolean keybint;
        boolean dosbanner;
        boolean allowvideoportaccess;
+       boolean emuretrace;
        boolean rdtsc;
        boolean timers;
        boolean mouse_flag;
@@ -331,12 +349,30 @@ typedef struct vesamode_type_struct {
        unsigned short detach;
        unsigned char *debugout;
        unsigned char *pre_stroke;  /* pointer to keyboard pre strokes */
+       unsigned char *pre_stroke_mem;
 
        /* Lock File business */
        char *tty_lockdir;	/* The Lock directory  */
        char *tty_lockfile;	/* Lock file pretext ie LCK.. */
        boolean tty_lockbinary;	/* Binary lock files ? */
 
+       /* list of arbitrary features
+        * (at minimum 1, will be increased when needed)
+        * The purpose of these parameters is to switch between code
+        * or code pieces that in principal do the same, but for some
+        * unknown reasons behave different between machines.
+        * If a 'features' becomes obsolete (problem solved) it will
+        * remain dummy for a while before re-used.
+        *
+        * NOTE: 'features' are not subject to permanent documentation!
+        *       They should be considered 'temporary hacks'
+        *
+        * Currently assigned:
+        *
+        *   features[0]		use new(0) or old(1) Slangcode
+        */
+       int features[1];
+
        /* Sound emulation */
        __u16 sb_base;
        __u8 sb_dma;
@@ -416,6 +452,7 @@ EXTERN void sigquit(int);
 #ifdef __linux__
 EXTERN void sigalrm(int, struct sigcontext_struct);
 EXTERN void sigio(int, struct sigcontext_struct);
+EXTERN int dosemu_sigaction(int sig, struct sigaction *new, struct sigaction *old);
 #endif
 #ifdef __NetBSD__
 EXTERN void sigalrm(int, int, struct sigcontext *);
@@ -486,7 +523,7 @@ do { \
 			sigaction(sig, &sa, NULL);
 #endif
 #ifdef __linux__
-#ifdef __GLIBC__
+#if GLIBC_VERSION_CODE
 #define SignalHandler __sighandler_t
 #endif
 #define NEWSETQSIG(sig, fun)	sa.sa_handler = (__sighandler_t)fun; \
@@ -563,6 +600,6 @@ EXTERN void read_next_scancode_from_queue (void);
 EXTERN unsigned short detach (void);
 EXTERN void HMA_init(void);
 
-extern char path_cdrom[];
+extern char *Path_cdrom[];
 
 #endif /* EMU_H */
