@@ -126,6 +126,12 @@
  */
 #define INT10_WATCHER_SEG	ROMBIOSSEG
 #define INT10_WATCHER_OFF	0x4240
+#ifdef X86_EMULATOR
+#define CPUEMU_WATCHER_SEG	ROMBIOSSEG
+#define CPUEMU_WATCHER_OFF	0x42A0
+#define CPUEMUI10_ADD		((CPUEMU_WATCHER_SEG << 4) +\
+				  CPUEMU_WATCHER_OFF + 11)
+#endif
 
 /* This inline interrupt is used for FCB open calls */
 #define INTE7_SEG	ROMBIOSSEG
@@ -136,6 +142,10 @@
 #define PIC_OFF       0x47f0
 #define PIC_ADD       ((PIC_SEG << 4) + PIC_OFF)
 
+#define CBACK_SEG	ROMBIOSSEG	/* callback return to dosemu */
+#define CBACK_OFF	0x47f1
+#define CBACK_ADD	((CBACK_SEG << 4) + CBACK_OFF)
+
 #define DPMI_SEG	ROMBIOSSEG
 #define DPMI_OFF	0x4800		/* need at least 512 bytes */
 #define DPMI_ADD	((DPMI_SEG << 4) + DPMI_OFF)
