@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -50,6 +50,9 @@ extern void pit_outp(ioport_t, Bit8u);
 #define PIT_TICK_RATE_M		1.193182
 #define PIT_TICK_RATE_f		0.193182
 
+/* number of 18.2Hz ticks = 86400*PIT_TICK_RATE/65536 = 1573042.6756 */
+#define TICKS_IN_A_DAY		1573042
+
 /*
  * Tick<->us conversion formulas:
  *	611/512		= 1.1933594	+150ppm
@@ -115,19 +118,30 @@ static __inline__ hitimer_t _mul64x32_(hitimer_t v, unsigned long f)
 {
 	hitimer_u t;
 
-	__asm__ __volatile__ ("
-		movl	%2,%%eax
-		mull	%%ebx
+#ifdef ASM_PEDANTIC
+	unsigned long int d0;
+#endif
+	__asm__ __volatile__ (
+#ifndef ASM_PEDANTIC
+	       "movl	%2,%%eax;"
+#endif
+	       "mull	%%ebx
 		movl	%%edx,%%ecx
 		movl	%3,%%eax
 		mull	%%ebx
 		addl	%%ecx,%%eax
 		adcl	$0,%%edx
 		"
+#ifndef ASM_PEDANTIC
 		: "=a"(t.t.tl),"=d"(t.t.th)
 		: "g"(((int *)&v)[0]),"g"(((int *)&v)[1]),\
 		  "b"(f)
 		: "%eax","%edx","%ecx","%ebx","memory" );
+#else
+		:"=&a"(t.t.tl),"=&d"(t.t.th), "=&b"(d0)
+		:"g"(((int *)&v)[1]),"2"(f),"0"(((int *)&v)[0])
+		:"%ecx","memory" );
+#endif
 	return t.td;
 }
 
@@ -154,6 +168,14 @@ static __inline__ hitimer_t _mul64x32_(hitimer_t v, unsigned long f)
 
 EXTERN hitimer_t (*GETcpuTIME)(void) INIT(0);
 
+static inline unsigned long long GETTSC(void) {
+	unsigned long long d;
+	__asm__ __volatile__ ("rdtsc" : "=A" (d));
+	return d;
+}
+
+#ifndef ASM_PEDANTIC
+
 #define CPUtoUS() \
 ({ \
 	hitimer_u t; \
@@ -195,6 +217,51 @@ EXTERN hitimer_t (*GETcpuTIME)(void) INIT(0);
 	t.td; \
 })
 
+#else /* ASM_PEDANTIC */
+
+#define CPUtoUS() \
+({ \
+	hitimer_u t; \
+	__asm__ __volatile__ (" \
+		rdtsc\n \
+		movl	%%edx,%%ebx\n \
+		mull	%2\n \
+		movl	%%edx,%%ecx\n \
+		movl	%%ebx,%%eax\n \
+		mull	%2\n \
+		addl	%%ecx,%%eax\n \
+		adcl	$0,%%edx\n \
+		" \
+		: "=&a"(t.t.tl),"=&d"(t.t.th) \
+		: "m"(config.cpu_spd) \
+		: "%ecx","%ebx","memory" ); \
+	t.td; \
+})
+
+#define CPUtoUS_Z() \
+({ \
+	hitimer_u t; \
+	__asm__ __volatile__ (" \
+		rdtsc\n \
+		movl	%%edx,%%ebx\n \
+		mull	%2\n \
+		movl	%%edx,%%ecx\n \
+		movl	%%ebx,%%eax\n \
+		mull	%2\n \
+		addl	%%ecx,%%eax\n \
+		adcl	$0,%%edx\n \
+		subl	%3,%%eax\n \
+		sbbl	%4,%%edx\n \
+		" \
+		: "=&a"(t.t.tl),"=&d"(t.t.th) \
+		: "m"(config.cpu_spd), \
+		  "m"(ZeroTimeBase.t.tl),"m"(ZeroTimeBase.t.th) \
+		: "%ecx","%ebx","memory" ); \
+	t.td; \
+})
+
+#endif /* ASM_PEDANTIC */
+
 /* 1 us granularity */
 extern hitimer_t GETusTIME(int sc);
 extern hitimer_t GETusSYSTIME(void);
@@ -202,10 +269,9 @@ extern hitimer_t GETusSYSTIME(void);
 /* 838 ns granularity */
 extern hitimer_t GETtickTIME(int sc);
 
-int stop_cputime (void);
-int restart_cputime (void);
+int stop_cputime (int);
+int restart_cputime (int);
 extern int cpu_time_stop;	/* for dosdebug */
-int getmhz(void);		/* for init/config.c */
 int bogospeed(unsigned long *spus, unsigned long *sptick);
 
 /* --------------------------------------------------------------------- */
