@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -255,7 +255,7 @@ static int preprocess(FILE *f, FILE *fo)
     return 1;
   }
 
-  inline _fputc(char c) {
+  inline void _fputc(char c) {
     fputc(c,fo);
     if (copy) {
       cb[cbuf_i++] = c;
@@ -391,7 +391,7 @@ static int usage()
   exit(1);
 }
 
-main (int argc, char** argv)
+int main (int argc, char** argv)
 {
   if (argc <= 1) usage();
 
